from __future__ import annotations

import copy
import inspect
import logging
import os
import pathlib
import uuid
from pprint import pformat as pf
from typing import TYPE_CHECKING, Any, Callable, Type

import pytest
from pytest import MonkeyPatch, param

import great_expectations as gx
import great_expectations.execution_engine.pandas_execution_engine
from great_expectations.compatibility import pydantic
from great_expectations.datasource.fluent import PandasDatasource
from great_expectations.datasource.fluent.dynamic_pandas import PANDAS_VERSION
from great_expectations.datasource.fluent.interfaces import Batch
from great_expectations.datasource.fluent.pandas_datasource import (
    _DYNAMIC_ASSET_TYPES,
    CSVAsset,
    DataFrameAsset,
    TableAsset,
    _PandasDataAsset,
)
from great_expectations.datasource.fluent.sources import (
    DEFAULT_PANDAS_DATA_ASSET_NAME,
    DEFAULT_PANDAS_DATASOURCE_NAME,
    DefaultPandasDatasourceError,
    _get_field_details,
)
from great_expectations.exceptions.exceptions import BuildBatchRequestError
from great_expectations.execution_engine.pandas_batch_data import PandasBatchData
from great_expectations.util import camel_to_snake

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.data_context import AbstractDataContext


logger = logging.getLogger(__file__)

# apply markers to entire test module
pytestmark = [
    pytest.mark.skipif(
        PANDAS_VERSION < 1.2, reason=f"Fluent pandas not supported on {PANDAS_VERSION}"
    )
]


@pytest.fixture
def pandas_datasource() -> PandasDatasource:
    return PandasDatasource(
        name="pandas_datasource",
    )


@pytest.fixture
def valid_file_path(csv_path: pathlib.Path) -> pathlib.Path:
    return csv_path / "yellow_tripdata_sample_2018-03.csv"


class SpyInterrupt(RuntimeError):
    """
    Exception that may be raised to interrupt the control flow of the program
    when a spy has already captured everything needed.
    """


@pytest.fixture
def capture_reader_fn_params(monkeypatch: MonkeyPatch):
    """
    Capture the `reader_options` arguments being passed to the `PandasExecutionEngine`.

    Note this fixture is heavily reliant on the implementation details of `PandasExecutionEngine`,
    should this change this fixture will need to change.
    """
    captured_args: list[list] = []
    captured_kwargs: list[dict[str, Any]] = []

    def reader_fn_spy(*args, **kwargs):
        logging.info(f"reader_fn_spy() called with...\n{args}\n{kwargs}")
        captured_args.append(args)
        captured_kwargs.append(kwargs)
        raise SpyInterrupt("Reader options have been captured")

    monkeypatch.setattr(
        great_expectations.execution_engine.pandas_execution_engine.PandasExecutionEngine,
        "_get_reader_fn",
        lambda *_: reader_fn_spy,
        raising=True,
    )

    yield captured_args, captured_kwargs


@pytest.mark.unit
class TestDynamicPandasAssets:
    @pytest.mark.parametrize(
        "method_name",
        [
            param("read_clipboard"),
            param("read_csv"),
            param("read_excel"),
            param("read_feather"),
            param("read_fwf"),
            param("read_gbq"),
            param("read_hdf"),
            param("read_html"),
            param("read_json"),
            param("read_orc"),
            param("read_parquet"),
            param("read_pickle"),
            param("read_sas"),
            param("read_spss"),
            param("read_sql"),
            param("read_sql_query"),
            param("read_sql_table"),
            param("read_stata"),
            param("read_table"),
            param(
                "read_xml",
                marks=pytest.mark.skipif(
                    PANDAS_VERSION < 1.3,
                    reason=f"read_xml does not exist on {PANDAS_VERSION} ",
                ),
            ),
        ],
    )
    def test_data_asset_defined_for_io_read_method(self, method_name: str):
        _, type_name = method_name.split("read_")
        assert type_name

        asset_class_names: set[str] = {
            camel_to_snake(t.__name__).split("_asset")[0] for t in PandasDatasource.asset_types
        }

        assert type_name in PandasDatasource._type_lookup
        assert type_name in asset_class_names

    @pytest.mark.parametrize("asset_class", _DYNAMIC_ASSET_TYPES)
    def test_add_asset_method_exists_and_is_functional(self, asset_class: Type[_PandasDataAsset]):
        type_name: str = _get_field_details(asset_class, "type").default_value
        method_name: str = f"add_{type_name}_asset"

        print(f"{method_name}() -> {asset_class.__name__}")

        assert method_name in PandasDatasource.__dict__

        ds = PandasDatasource(
            name="ds_for_testing_add_asset_methods",
        )
        method = getattr(ds, method_name)

        with pytest.raises(pydantic.ValidationError) as exc_info:
            positional_arg_string = "foo"
            positional_args: list[str] = []
            while len(positional_args) < 3:
                try:
                    method(
                        f"{asset_class.__name__}_add_asset_test",
                        *positional_args,
                        _invalid_key="bar",
                    )
                    break
                except TypeError:
                    positional_args.append(positional_arg_string)
        # importantly check that the method creates (or attempts to create) the intended asset
        assert exc_info.value.model == asset_class

    @pytest.mark.parametrize("asset_class", _DYNAMIC_ASSET_TYPES)
    def test_add_asset_method_signature(self, asset_class: Type[_PandasDataAsset]):
        type_name: str = _get_field_details(asset_class, "type").default_value
        method_name: str = f"add_{type_name}_asset"

        ds = PandasDatasource(
            name="ds_for_testing_add_asset_methods",
        )
        method = getattr(ds, method_name)

        add_asset_method_sig: inspect.Signature = inspect.signature(method)
        print(f"\t{method_name}()\n{add_asset_method_sig}\n")

        asset_class_init_sig: inspect.Signature = inspect.signature(asset_class)
        print(f"\t{asset_class.__name__}\n{asset_class_init_sig}\n")

        for i, param_name in enumerate(asset_class_init_sig.parameters):
            print(f"{i} {param_name} ", end="")

            if param_name == "type":
                assert param_name not in add_asset_method_sig.parameters, (
                    "type should not be part of the `add_<TYPE>_asset` method"
                )
                print("⏩")
                continue

            assert param_name in add_asset_method_sig.parameters
            print("✅")

    @pytest.mark.parametrize("asset_class", _DYNAMIC_ASSET_TYPES)
    def test_minimal_validation(self, asset_class: Type[_PandasDataAsset]):
        """
        These parametrized tests ensures that every `PandasDatasource` asset model does some minimal
        validation, and doesn't accept arbitrary keyword arguments.
        This is also a proxy for testing that the dynamic pydantic model creation was successful.
        """
        with pytest.raises(pydantic.ValidationError) as exc_info:
            asset_class(  # type: ignore[call-arg] # FIXME CoP
                name="test",
                invalid_keyword_arg="bad",
            )

        errors_dict = exc_info.value.errors()
        assert errors_dict[  # the extra keyword error will always be the last error
            -1  # we don't care about any other errors for this test
        ] == {
            "loc": ("invalid_keyword_arg",),
            "msg": "extra fields not permitted",
            "type": "value_error.extra",
        }

    @pytest.mark.parametrize(
        ["asset_model", "extra_kwargs"],
        [
            (CSVAsset, {"sep": "|", "names": ["col1", "col2", "col3"]}),
            (
                TableAsset,
                {
                    "sep": "|",
                    "names": ["col1", "col2", "col3", "col4"],
                    "skiprows": [2, 4, 5],
                },
            ),
        ],
    )
    def test_data_asset_defaults(
        self,
        csv_path: pathlib.Path,
        asset_model: Type[_PandasDataAsset],
        extra_kwargs: dict,
    ):
        """
        Test that an asset dictionary can be dumped with only the original passed keys
        present.
        """
        kwargs: dict[str, Any] = {
            "name": "test",
            "filepath_or_buffer": csv_path / "yellow_tripdata_sample_2018-04.csv",
        }
        kwargs.update(extra_kwargs)
        print(f"extra_kwargs\n{pf(extra_kwargs)}")
        asset_instance = asset_model(**kwargs)
        assert asset_instance.dict(exclude={"type"}) == kwargs

    @pytest.mark.parametrize(
        "extra_kwargs",
        [
            {"sep": "|", "decimal": ","},
            {"usecols": [0, 1, 2], "names": ["foo", "bar"]},
            {"dtype": {"col_1": "Int64"}},
        ],
    )
    def test_data_asset_reader_options_passthrough(
        self,
        empty_data_context: AbstractDataContext,
        csv_path: pathlib.Path,
        capture_reader_fn_params: tuple[list[list], list[dict]],
        extra_kwargs: dict,
    ):
        extra_kwargs.update({"filepath_or_buffer": csv_path / "yellow_tripdata_sample_2018-04.csv"})
        batch_request = (
            empty_data_context.data_sources.add_pandas(
                "my_pandas",
            )
            .add_csv_asset(
                "my_csv",
                **extra_kwargs,
            )
            .build_batch_request()
        )
        with pytest.raises(SpyInterrupt):
            empty_data_context.get_validator(batch_request=batch_request)

        captured_args, captured_kwargs = capture_reader_fn_params
        print(f"positional args:\n{pf(captured_args[-1])}\n")
        print(f"keyword args:\n{pf(captured_kwargs[-1])}")

        assert captured_kwargs[-1] == extra_kwargs

    @pytest.mark.parametrize(
        "read_method_name,positional_args",
        [
            param("read_clipboard", {}),
            param("read_csv", {"filepath_or_buffer": "valid_file_path"}),
            param("read_excel", {"io": "valid_file_path"}),
            param("read_feather", {"path": "valid_file_path"}),
            param("read_fwf", {"filepath_or_buffer": "valid_file_path"}),
            param("read_gbq", {"query": "SELECT * FROM my_table"}),
            param("read_hdf", {"path_or_buf": "valid_file_path"}),
            param("read_html", {"io": "valid_file_path"}),
            param("read_json", {"path_or_buf": "valid_file_path"}),
            param("read_orc", {"path": "valid_file_path"}),
            param("read_parquet", {"path": "valid_file_path"}),
            param("read_pickle", {"filepath_or_buffer": "valid_file_path"}),
            param("read_sas", {"filepath_or_buffer": "valid_file_path"}),
            param("read_spss", {"path": "valid_file_path"}),
            param("read_sql", {"sql": "SELECT * FROM my_table", "con": "sqlite://"}),
            param("read_sql_query", {"sql": "SELECT * FROM my_table", "con": "sqlite://"}),
            param("read_sql_table", {"table_name": "my_table", "con": "sqlite://"}),
            param("read_stata", {"filepath_or_buffer": "valid_file_path"}),
            param("read_table", {"filepath_or_buffer": "valid_file_path"}),
            param(
                "read_xml",
                {"path_or_buffer": "valid_file_path"},
                marks=pytest.mark.skipif(
                    PANDAS_VERSION < 1.3,
                    reason=f"read_xml does not exist on {PANDAS_VERSION} ",
                ),
            ),
        ],
    )
    def test_positional_arguments(
        self,
        mocker,
        empty_data_context: AbstractDataContext,
        read_method_name: str,
        positional_args: dict[str, str | pathlib.Path],
        request,
    ):
        if "valid_file_path" in positional_args.values():
            positional_args = {
                positional_arg_name: request.getfixturevalue("valid_file_path")
                for positional_arg_name, positional_arg in positional_args.items()
                if positional_arg == "valid_file_path"
            }

        add_method_name = "add_" + read_method_name.split("read_")[1] + "_asset"
        add_method: Callable = getattr(
            empty_data_context.data_sources.pandas_default, add_method_name
        )

        asset: _PandasDataAsset = add_method(
            "my_asset",
            *positional_args.values(),
        )
        for positional_arg_name, positional_arg in positional_args.items():
            assert getattr(asset, positional_arg_name) == positional_arg

        read_method: Callable = getattr(
            empty_data_context.data_sources.pandas_default, read_method_name
        )
        # This is not a an ideal mock.
        # In this test we are validating that the read_method for a particular pandas datasource
        # has the correct positional arguments.
        # We don't care about the actual data being read in and the batch that will be produced from that data.  # noqa: E501 # FIXME CoP
        # In fact, we call all our read methods on a path which might not be readable by the reader (eg calling  # noqa: E501 # FIXME CoP
        # read_json on a csv file). We patch the internal call that actually tries to read and create the batch.  # noqa: E501 # FIXME CoP
        # Ideally, we would rewrite this test so we wouldn't need to mock like this.
        mocker.patch(
            "great_expectations.datasource.fluent.pandas_datasource._PandasDataAsset.get_batch"
        )
        # read_* normally returns batch but, since we've added a mock in the line above, we get a mock object returned.  # noqa: E501 # FIXME CoP
        # We are calling it here for it's side effect on the default asset so get and inspect that afterwards.  # noqa: E501 # FIXME CoP
        _ = read_method(*positional_args.values())
        default_asset = empty_data_context.data_sources.pandas_default.get_asset(
            name=DEFAULT_PANDAS_DATA_ASSET_NAME
        )
        for positional_arg_name, positional_arg in positional_args.items():
            assert getattr(default_asset, positional_arg_name) == positional_arg


@pytest.mark.filesystem
def test_default_pandas_datasource_get_and_set(
    empty_data_context: AbstractDataContext, valid_file_path: pathlib.Path
):
    pandas_datasource = empty_data_context.data_sources.pandas_default
    assert isinstance(pandas_datasource, PandasDatasource)
    assert pandas_datasource.name == DEFAULT_PANDAS_DATASOURCE_NAME
    assert len(pandas_datasource.assets) == 0

    batch = pandas_datasource.read_csv(
        filepath_or_buffer=valid_file_path,
    )
    assert isinstance(batch, Batch)
    csv_data_asset_1 = pandas_datasource.get_asset(name=DEFAULT_PANDAS_DATA_ASSET_NAME)
    assert isinstance(csv_data_asset_1, _PandasDataAsset)
    assert csv_data_asset_1.name == DEFAULT_PANDAS_DATA_ASSET_NAME
    assert len(pandas_datasource.assets) == 1

    # ensure we get the same datasource when we call pandas_default again
    pandas_datasource = empty_data_context.data_sources.pandas_default
    assert pandas_datasource.name == DEFAULT_PANDAS_DATASOURCE_NAME
    assert len(pandas_datasource.assets) == 1
    assert pandas_datasource.get_asset(name=DEFAULT_PANDAS_DATA_ASSET_NAME)

    # ensure we overwrite the ephemeral data asset if no name is passed
    _ = pandas_datasource.read_csv(filepath_or_buffer=valid_file_path)
    assert csv_data_asset_1.name == DEFAULT_PANDAS_DATA_ASSET_NAME
    assert len(pandas_datasource.assets) == 1

    # ensure we get an additional named asset when one is passed
    expected_csv_data_asset_name = "my_csv_asset"
    _ = pandas_datasource.read_csv(
        asset_name=expected_csv_data_asset_name,
        filepath_or_buffer=valid_file_path,
    )
    csv_data_asset_2 = pandas_datasource.get_asset(name=expected_csv_data_asset_name)
    assert csv_data_asset_2.name == expected_csv_data_asset_name
    assert len(pandas_datasource.assets) == 2

    # ensure ephemeral data assets are not serialized
    config_as_dict = empty_data_context.fluent_config.dict()["fluent_datasources"]
    print(f"{pf(config_as_dict)}")
    for ds in config_as_dict:
        for asset in ds.get("assets", []):
            assert asset["name"] != DEFAULT_PANDAS_DATA_ASSET_NAME


@pytest.mark.spark
def test_default_pandas_datasource_name_conflict(
    empty_data_context: AbstractDataContext,
):
    # empty_data_context.data_sources.add_spark(name=DEFAULT_PANDAS_DATASOURCE_NAME)
    empty_data_context.data_sources.add_spark(name=DEFAULT_PANDAS_DATASOURCE_NAME)
    with pytest.raises(DefaultPandasDatasourceError):
        _ = empty_data_context.data_sources.pandas_default

    # the datasource name is available
    empty_data_context.data_sources.all().pop(DEFAULT_PANDAS_DATASOURCE_NAME)
    pandas_datasource = empty_data_context.data_sources.pandas_default
    assert isinstance(pandas_datasource, PandasDatasource)
    assert pandas_datasource.name == DEFAULT_PANDAS_DATASOURCE_NAME


@pytest.mark.filesystem
def test_read_dataframe(empty_data_context: AbstractDataContext, test_df_pandas: pd.DataFrame):
    # validates that a dataframe object is passed
    with pytest.raises(ValueError) as exc_info:
        _ = empty_data_context.data_sources.pandas_default.read_dataframe(dataframe={})  # type: ignore[arg-type] # FIXME CoP

    assert (
        'Cannot execute "PandasDatasource.read_dataframe()" without a valid "dataframe" argument.'
        in str(exc_info.value)
    )

    # correct working behavior with read method
    datasource = empty_data_context.data_sources.pandas_default
    batch = datasource.read_dataframe(dataframe=test_df_pandas)
    assert isinstance(batch, Batch)
    assert isinstance(
        empty_data_context.data_sources.pandas_default.get_asset(
            name=DEFAULT_PANDAS_DATA_ASSET_NAME
        ),
        DataFrameAsset,
    )

    # correct working behavior with add method
    dataframe_asset_name = "my_dataframe_asset"
    dataframe_asset = empty_data_context.data_sources.pandas_default.add_dataframe_asset(
        name=dataframe_asset_name
    )
    assert isinstance(dataframe_asset, DataFrameAsset)
    assert dataframe_asset.name == "my_dataframe_asset"
    assert len(empty_data_context.data_sources.pandas_default.assets) == 2
    bd = dataframe_asset.add_batch_definition_whole_dataframe(name="bd")
    bd_batch = bd.get_batch(batch_parameters={"dataframe": test_df_pandas})
    for b in [batch, bd_batch]:
        assert isinstance(b.data, PandasBatchData)
        b.data.dataframe.equals(test_df_pandas)


@pytest.mark.cloud
def test_cloud_get_csv_asset_not_in_memory(valid_file_path: pathlib.Path):
    # this test runs end-to-end in a real Cloud Data Context
    context = gx.get_context(
        mode="cloud",
        cloud_base_url=os.environ.get("GX_CLOUD_BASE_URL"),
        cloud_organization_id=os.environ.get("GX_CLOUD_ORGANIZATION_ID"),
        cloud_access_token=os.environ.get("GX_CLOUD_ACCESS_TOKEN"),
    )
    datasource_name = f"DS_{uuid.uuid4().hex}"
    csv_asset_name = f"DA_{uuid.uuid4().hex}"
    datasource = context.data_sources.add_pandas(name=datasource_name)
    try:
        _ = datasource.add_csv_asset(
            name=csv_asset_name,
            filepath_or_buffer=valid_file_path,
        )
        csv_asset = datasource.get_asset(name=csv_asset_name)
        csv_asset.build_batch_request()

        assert csv_asset_name not in context.data_sources.all()._in_memory_data_assets
    finally:
        context.data_sources.delete(name=datasource_name)


@pytest.mark.filesystem
def test_pandas_data_asset_batch_metadata(
    empty_data_context: AbstractDataContext, valid_file_path: pathlib.Path
):
    my_config_variables = {"pipeline_filename": __file__}
    empty_data_context.config_variables.update(my_config_variables)

    pandas_datasource = empty_data_context.data_sources.pandas_default

    batch_metadata = {
        "no_curly_pipeline_filename": "$pipeline_filename",
        "curly_pipeline_filename": "${pipeline_filename}",
        "pipeline_step": "transform_3",
    }

    csv_asset = pandas_datasource.add_csv_asset(
        name="my_csv_asset",
        filepath_or_buffer=valid_file_path,
        batch_metadata=batch_metadata,
    )
    assert csv_asset.batch_metadata == batch_metadata

    batch = csv_asset.get_batch(csv_asset.build_batch_request())

    # allow mutation of this attribute
    batch.metadata["also_this_one"] = "other_batch-level_value"

    substituted_batch_metadata = copy.deepcopy(batch_metadata)
    substituted_batch_metadata.update(
        {
            "no_curly_pipeline_filename": __file__,
            "curly_pipeline_filename": __file__,
            "also_this_one": "other_batch-level_value",
        }
    )
    assert batch.metadata == substituted_batch_metadata


@pytest.mark.filesystem
def test_build_batch_request_raises_if_missing_dataframe(
    empty_data_context: AbstractDataContext,
):
    dataframe_asset = empty_data_context.data_sources.add_or_update_pandas(
        name="fluent_pandas_datasource"
    ).add_dataframe_asset(name="my_df_asset")

    with pytest.raises(BuildBatchRequestError) as e:
        dataframe_asset.build_batch_request()

    assert str(e.value).startswith("Bad input to build_batch_request:")



from logging import Logger
from typing import ClassVar, List, Type

from great_expectations.datasource.fluent.data_asset.path.file_asset import FileDataAsset
from great_expectations.datasource.fluent.interfaces import DataAsset as DataAsset
from great_expectations.datasource.fluent.pandas_datasource import _PandasDatasource

logger: Logger

class _PandasFilePathDatasource(_PandasDatasource):
    asset_types: ClassVar[List[Type[DataAsset]]]
    assets: List[FileDataAsset]



from __future__ import annotations

import hashlib
import random
from typing import TYPE_CHECKING

import great_expectations.exceptions as gx_exceptions
from great_expectations.execution_engine.partition_and_sample.data_sampler import (
    DataSampler,
)

if TYPE_CHECKING:
    import pandas as pd

    from great_expectations.core.id_dict import BatchSpec


class PandasDataSampler(DataSampler):
    """Methods for sampling a pandas dataframe."""

    def sample_using_limit(self, df: pd.DataFrame, batch_spec: BatchSpec) -> pd.DataFrame:
        """Sample the first n rows of data.

        Args:
            df: pandas dataframe.
            batch_spec: Should contain key `n` in sampling_kwargs, the number of
                values in the sample e.g. sampling_kwargs={"n": 100}.

        Returns:
            Sampled dataframe

        Raises:
            SamplerError
        """
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("n", batch_spec)

        n: int = batch_spec["sampling_kwargs"]["n"]
        return df.head(n)

    def sample_using_random(
        self,
        df: pd.DataFrame,
        batch_spec: BatchSpec,
    ) -> pd.DataFrame:
        """Take a random sample of rows, retaining proportion p.

        Args:
            df: dataframe to sample
            batch_spec: Can contain key `p` (float) which defaults to 0.1
                if not provided.

        Returns:
            Sampled dataframe

        Raises:
            SamplerError
        """
        p: float = self.get_sampling_kwargs_value_or_default(
            batch_spec=batch_spec, sampling_kwargs_key="p", default_value=0.1
        )
        return df[df.index.map(lambda x: random.random() < p)]

    def sample_using_mod(
        self,
        df: pd.DataFrame,
        batch_spec: BatchSpec,
    ) -> pd.DataFrame:
        """Take the mod of named column, and only keep rows that match the given value.

        Args:
            df: dataframe to sample
            batch_spec: should contain keys `column_name`, `mod` and `value`

        Returns:
            Sampled dataframe

        Raises:
            SamplerError
        """
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("column_name", batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("mod", batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("value", batch_spec)
        column_name: str = self.get_sampling_kwargs_value_or_default(batch_spec, "column_name")
        mod: int = self.get_sampling_kwargs_value_or_default(batch_spec, "mod")
        value: int = self.get_sampling_kwargs_value_or_default(batch_spec, "value")

        return df[df[column_name].map(lambda x: x % mod == value)]

    def sample_using_a_list(
        self,
        df: pd.DataFrame,
        batch_spec: BatchSpec,
    ) -> pd.DataFrame:
        """Match the values in the named column against value_list, and only keep the matches.

        Args:
            df: dataframe to sample
            batch_spec: should contain keys `column_name` and `value_list`

        Returns:
            Sampled dataframe

        Raises:
            SamplerError
        """
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("column_name", batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("value_list", batch_spec)

        column_name: str = self.get_sampling_kwargs_value_or_default(batch_spec, "column_name")
        value_list: int = self.get_sampling_kwargs_value_or_default(batch_spec, "value_list")

        return df[df[column_name].isin(value_list)]  # type: ignore[arg-type] # FIXME CoP

    def sample_using_hash(
        self,
        df: pd.DataFrame,
        batch_spec: BatchSpec,
    ) -> pd.DataFrame:
        """Hash the values in the named column, and only keep rows that match the given hash_value.

        Args:
            df: dataframe to sample
            batch_spec: should contain keys `column_name` and optionally `hash_digits`
                (default is 1 if not provided), `hash_value` (default is "f" if not provided),
                and `hash_function_name` (default is "md5" if not provided)

        Returns:
            Sampled dataframe

        Raises:
            SamplerError
        """
        self.verify_batch_spec_sampling_kwargs_exists(batch_spec)
        self.verify_batch_spec_sampling_kwargs_key_exists("column_name", batch_spec)
        column_name: str = self.get_sampling_kwargs_value_or_default(batch_spec, "column_name")
        hash_digits: int = self.get_sampling_kwargs_value_or_default(
            batch_spec=batch_spec, sampling_kwargs_key="hash_digits", default_value=1
        )
        hash_value: str = self.get_sampling_kwargs_value_or_default(
            batch_spec=batch_spec, sampling_kwargs_key="hash_value", default_value="f"
        )

        hash_function_name: str = self.get_sampling_kwargs_value_or_default(
            batch_spec=batch_spec,
            sampling_kwargs_key="hash_function_name",
            default_value="md5",
        )

        try:
            hash_func = getattr(hashlib, hash_function_name)
        except (TypeError, AttributeError):
            raise (
                gx_exceptions.ExecutionEngineError(  # noqa: TRY003 # FIXME CoP
                    f"""The sampling method used with PandasExecutionEngine has a reference to an invalid hash_function_name.
                       Reference to {hash_function_name} cannot be found."""  # noqa: E501 # FIXME CoP
                )
            )

        matches: pd.Series = df[column_name].map(
            lambda x: hash_func(str(x).encode()).hexdigest()[-1 * hash_digits :] == hash_value
        )
        return df[matches]



from __future__ import annotations

import logging
from abc import ABC
from typing import TYPE_CHECKING, Optional

from great_expectations.exceptions.exceptions import (
    InvalidExpectationConfigurationError,
)
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.expectation import (
    ColumnMapExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metrics.map_metric_provider import (
    ColumnMapMetricProvider,
    column_condition_partial,
)
from great_expectations.render import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import (
    num_to_str,
    parse_row_condition_string_pandas_engine,
    substitute_none_for_missing,
)
from great_expectations.util import camel_to_snake

if TYPE_CHECKING:
    from great_expectations.core import (
        ExpectationValidationResult,
    )
    from great_expectations.expectations.expectation_configuration import (
        ExpectationConfiguration,
    )
    from great_expectations.render.renderer_configuration import AddParamArgs

logger = logging.getLogger(__name__)


class SetColumnMapMetricProvider(ColumnMapMetricProvider):
    """Base class for all SetColumnMapMetrics.

    SetColumnMapMetric classes inheriting from SetColumnMapMetricProvider are ephemeral,
    defined by their `set` attribute, and registered during the execution of their associated SetColumnMapExpectation.

    Metric Registration Example:

    ```python
    map_metric = SetBasedColumnMapExpectation.register_metric(
        set_camel_name='SolfegeScale',
        set_=['do', 're', 'mi', 'fa', 'so', 'la', 'ti'],
    )
    ```

    In some cases, subclasses of MetricProvider, such as SetColumnMapMetricProvider, will already
    have correct values that may simply be inherited by Metric classes.

    Args:
        set_ (union[list, set]): A value set.
        metric_name (str): The name of the registered metric. Must be globally unique in a great_expectations installation.
            Constructed by the `register_metric(...)` function during Expectation execution.
        domain_keys (tuple): A tuple of the keys used to determine the domain of the metric.
        condition_value_keys (tuple): A tuple of the keys used to determine the value of the metric.
    """  # noqa: E501 # FIXME CoP

    condition_value_keys = ()

    @column_condition_partial(engine=PandasExecutionEngine)
    def _pandas(cls, column, **kwargs):
        return column.isin(cls.set_)

    @column_condition_partial(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(cls, column, _dialect, **kwargs):
        return column.in_(cls.set_)

    @column_condition_partial(engine=SparkDFExecutionEngine)
    def _spark(cls, column, **kwargs):
        return column.isin(cls.set_)


class SetBasedColumnMapExpectation(ColumnMapExpectation, ABC):
    """Base class for SetBasedColumnMapExpectations.

    SetBasedColumnMapExpectations facilitate set-based comparisons as the core logic for a Map Expectation.

    Example Definition:

    ```python
    ExpectColumnValuesToBeInSolfegeScaleSet(SetBasedColumnMapExpectation):
        set_camel_name = SolfegeScale
        set_ = ['do', 're', 'mi', 'fa', 'so', 'la', 'ti']
        set_semantic_name = "the Solfege scale"
        map_metric = SetBasedColumnMapExpectation.register_metric(
            set_camel_name=set_camel_name,
            set_=set_
    )
    ```

    Args:
        set_camel_name (str): A name describing a set of values, in camel case.
        set_ (str): A value set.
        set_semantic_name (optional[str]): A name for the semantic type representing the set being validated..
        map_metric (str): The name of an ephemeral metric, as returned by `register_metric(...)`.

    ---Documentation---
        - https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_set_based_column_map_expectations
    """  # noqa: E501 # FIXME CoP

    @staticmethod
    def register_metric(
        set_camel_name: str,
        set_: str,
    ) -> str:
        """Register an ephemeral metric using a constructed name with the logic provided by SetColumnMapMetricProvider.

        Args:
            set_camel_name: A name describing a set of values, in camel case.
            set_: A value set.

        Returns:
            map_metric: The constructed name of the ephemeral metric.
        """  # noqa: E501 # FIXME CoP
        set_snake_name = camel_to_snake(set_camel_name)
        map_metric: str = "column_values.match_" + set_snake_name + "_set"

        # Define the class using `type`. This allows us to name it dynamically.
        new_column_set_metric_provider = type(  # noqa: F841 # never used
            f"(ColumnValuesMatch{set_camel_name}Set",
            (SetColumnMapMetricProvider,),
            {
                "condition_metric_name": map_metric,
                "set_": set_,
            },
        )

        return map_metric

    def validate_configuration(
        self, configuration: Optional[ExpectationConfiguration] = None
    ) -> None:
        """Raise an exception if the configuration is not viable for an expectation.

        Args:
            configuration: An ExpectationConfiguration

        Raises:
            InvalidExpectationConfigurationError: If no `set_` or `column` specified, or if `mostly` parameter
                incorrectly defined.
        """  # noqa: E501 # FIXME CoP
        super().validate_configuration(configuration)
        try:
            assert getattr(self, "set_", None) is not None, (
                "set_ is required for SetBasedColumnMap Expectations"
            )

            assert "column" in configuration.kwargs, (
                "'column' parameter is required for ColumnMap expectations"
            )

            if "mostly" in configuration.kwargs:
                mostly = configuration.kwargs["mostly"]
                assert isinstance(mostly, (int, float)), (
                    "'mostly' parameter must be an integer or float"
                )
                assert 0 <= mostly <= 1, "'mostly' parameter must be between 0 and 1"

        except AssertionError as e:
            raise InvalidExpectationConfigurationError(str(e))

    # question, descriptive, prescriptive, diagnostic
    @classmethod
    @renderer(renderer_type=LegacyRendererType.QUESTION)
    def _question_renderer(cls, configuration, result=None, runtime_configuration=None):
        column = configuration.kwargs.get("column")
        mostly = configuration.kwargs.get("mostly")
        set_ = cls.set_
        set_semantic_name = getattr(cls, "set_semantic_name", None)

        if mostly == 1 or mostly is None:
            if set_semantic_name is not None:
                return f'Are all values in column "{column}" in {set_semantic_name}: {set_!s}?'
            else:
                return f'Are all values in column "{column}" in the set {set_!s}?'
        else:  # noqa: PLR5501 # FIXME CoP
            if set_semantic_name is not None:
                return f'Are at least {mostly * 100}% of values in column "{column}" in {set_semantic_name}: {set_!s}?'  # noqa: E501 # FIXME CoP
            else:
                return f'Are at least {mostly * 100}% of values in column "{column}" in the set {set_!s}?'  # noqa: E501 # FIXME CoP

    @classmethod
    @renderer(renderer_type=LegacyRendererType.ANSWER)
    def _answer_renderer(cls, configuration=None, result=None, runtime_configuration=None):
        column = result.expectation_config.kwargs.get("column")
        mostly = result.expectation_config.kwargs.get("mostly")
        set_ = cls.set_
        set_semantic_name = getattr(cls, "set_semantic_name", None)

        if result.success:
            if mostly == 1 or mostly is None:
                if set_semantic_name is not None:
                    return f'All values in column "{column}" are in {set_semantic_name}: {set_!s}.'
                else:
                    return f'All values in column "{column}" are in the set {set_!s}.'
            else:  # noqa: PLR5501 # FIXME CoP
                if set_semantic_name is not None:
                    return f'At least {mostly * 100}% of values in column "{column}" are in {set_semantic_name}: {set_!s}.'  # noqa: E501 # FIXME CoP
                else:
                    return f'At least {mostly * 100}% of values in column "{column}" are in the set {set!s}.'  # noqa: E501 # FIXME CoP
        else:  # noqa: PLR5501 # FIXME CoP
            if set_semantic_name is not None:
                return f' Less than {mostly * 100}% of values in column "{column}" are in {set_semantic_name}: {set_!s}.'  # noqa: E501 # FIXME CoP
            else:
                return f'Less than {mostly * 100}% of values in column "{column}" are in the set {set_!s}.'  # noqa: E501 # FIXME CoP

    @classmethod
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        add_param_args: AddParamArgs = (
            ("column", RendererValueType.STRING),
            ("mostly", RendererValueType.NUMBER),
            ("set_", RendererValueType.STRING),
            ("set_semantic_name", RendererValueType.STRING),
        )
        for name, param_type in add_param_args:
            renderer_configuration.add_param(name=name, param_type=param_type)

        params = renderer_configuration.params

        if not params.set_:
            template_str = "values must match a set but none was specified."
        else:
            if params.set_semantic_name:
                template_str = "values must match the set $set_semantic_name: $set_"
            else:
                template_str = "values must match this set: $set_"

            if params.mostly and params.mostly.value < 1.0:
                renderer_configuration = cls._add_mostly_pct_param(
                    renderer_configuration=renderer_configuration
                )
                template_str += ", at least $mostly_pct % of the time."
            else:
                template_str += "."

        if renderer_configuration.include_column_name:
            template_str = "$column " + template_str

        renderer_configuration.template_str = template_str

        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ):
        runtime_configuration = runtime_configuration or {}
        include_column_name = runtime_configuration.get("include_column_name") is not False
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,
            [
                "column",
                "set_",
                "mostly",
                "row_condition",
                "condition_parser",
                "set_semantic_name",
            ],
        )

        if not params.get("set_"):
            template_str = "values must match a set but none was specified."
        else:
            if params.get("set_semantic_name"):
                template_str = "values must match the set $set_semantic_name: $set_"
            else:
                template_str = "values must match this set: $set_"
            if params["mostly"] is not None:
                params["mostly_pct"] = num_to_str(params["mostly"] * 100, no_scientific=True)
                template_str += ", at least $mostly_pct % of the time."
            else:
                template_str += "."

        if include_column_name:
            template_str = "$column " + template_str

        if params["row_condition"] is not None:
            (
                conditional_template_str,
                conditional_params,
            ) = parse_row_condition_string_pandas_engine(params["row_condition"])
            template_str = conditional_template_str + ", then " + template_str
            params.update(conditional_params)

        params_with_json_schema = {  # noqa: F841 # never used
            "column": {"schema": {"type": "string"}, "value": params.get("column")},
            "mostly": {"schema": {"type": "number"}, "value": params.get("mostly")},
            "mostly_pct": {
                "schema": {"type": "number"},
                "value": params.get("mostly_pct"),
            },
            "set_": {"schema": {"type": "string"}, "value": params.get("set_")},
            "row_condition": {
                "schema": {"type": "string"},
                "value": params.get("row_condition"),
            },
            "condition_parser": {
                "schema": {"type": "string"},
                "value": params.get("condition_parser"),
            },
            "set_semantic_name": {
                "schema": {"type": "string"},
                "value": params.get("set_semantic_name"),
            },
        }

        return [
            RenderedStringTemplateContent(
                **{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": template_str,
                        "params": params,
                        "styling": styling,
                    },
                }
            )
        ]



from __future__ import annotations

import contextlib
import json
import logging
import re
import urllib.parse
import uuid
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Dict,
    Final,
    Generator,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    TypedDict,
    Union,
)

import responses

from great_expectations.compatibility import pydantic
from great_expectations.data_context.store.gx_cloud_store_backend import (
    ErrorDetail,
    ErrorPayload,
)
from great_expectations.datasource.fluent.config import GxConfig

if TYPE_CHECKING:
    from requests import PreparedRequest


LOGGER = logging.getLogger(__name__)

# ##########################
# Constants
# ##########################

MISSING: Final = object()

GX_CLOUD_MOCK_BASE_URL: Final[str] = "https://app.greatexpectations.fake.io/"

DUMMY_JWT_TOKEN: Final[str] = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
)
# Can replace hardcoded ids with dynamic ones if using a regex url with responses.add_callback()
# https://github.com/getsentry/responses/tree/master#dynamic-responses
FAKE_USER_ID: Final[str] = "00000000-0000-0000-0000-000000000000"
FAKE_ORG_ID: Final[str] = str(uuid.UUID("12345678123456781234567812345678"))
FAKE_DATA_CONTEXT_ID: Final[str] = str(uuid.uuid4())
UUID_REGEX: Final[str] = r"[a-f0-9-]{36}"

DEFAULT_HEADERS: Final[dict[str, str]] = {"content-type": "application/json"}

# ##########################
# Models & Types
# ##########################


class _DatasourceSchema(pydantic.BaseModel, extra="allow"):
    id: Optional[str] = None
    type: str
    name: str
    assets: List[dict] = pydantic.Field(default_factory=list)


class CloudResponseSchema(pydantic.BaseModel):
    data: _DatasourceSchema

    @classmethod
    def from_datasource_json(cls, ds_payload: str | bytes) -> CloudResponseSchema:
        data = json.loads(ds_payload)
        return cls(**data)


class CallbackResult(NamedTuple):
    status: int
    headers: dict[str, str]
    body: str


ErrorPayloadSchema = pydantic.create_model_from_typeddict(ErrorPayload)
ErrorPayloadSchema.update_forward_refs(ErrorDetail=ErrorDetail)


class CloudDetails(NamedTuple):
    base_url: str
    org_id: str
    access_token: str


FakeDBTypedDict = TypedDict(
    "FakeDBTypedDict",
    # using alternative syntax for creating type dict because of key names with hyphens
    # https://peps.python.org/pep-0589/#alternative-syntax
    {
        "me": Dict[str, str],
        "data-context-configuration": Dict[str, Union[str, dict]],
        "DATASOURCE_NAMES": Set[str],
        "datasources": Dict[str, dict],
        "EXPECTATION_SUITE_NAMES": Set[str],
        "expectation_suites": Dict[str, dict],
        "CHECKPOINT_NAMES": Set[str],
        "checkpoints": Dict[str, dict],
        "VALIDATION_DEFINITION_NAMES": Set[str],
        "validation_definitions": Dict[str, dict],
    },
)

# ##########################
# Helpers
# ##########################


@pydantic.validate_arguments
def create_fake_db_seed_data(fds_config: Optional[GxConfig] = None) -> FakeDBTypedDict:
    fds_config = fds_config or GxConfig(fluent_datasources=[])
    datasource_names: set[str] = set()
    datasource_config: dict[str, dict | str] = {}
    datasources_by_id: dict[str, dict] = {}

    for ds in fds_config._json_dict()["fluent_datasources"]:
        name: str = ds["name"]
        datasource_names.add(name)

        id: str = str(uuid.uuid4())
        ds["id"] = id

        ds_response_json = {"data": ds}

        datasource_config[name] = name
        datasources_by_id[id] = ds_response_json

    return {
        "me": {"user_id": FAKE_USER_ID},
        "DATASOURCE_NAMES": datasource_names,
        "datasources": datasources_by_id,
        "EXPECTATION_SUITE_NAMES": set(),
        "expectation_suites": {},
        "CHECKPOINT_NAMES": set(),
        "checkpoints": {},
        "data-context-configuration": {
            "anonymous_usage_statistics": {
                "data_context_id": FAKE_DATA_CONTEXT_ID,
                "enabled": False,
            },
            "datasources": datasource_config,
            "checkpoint_store_name": "default_checkpoint_store",
            "expectations_store_name": "default_expectations_store",
            "validation_results_store_name": "default_validation_results_store",
            "stores": {
                "default_expectations_store": {
                    "class_name": "ExpectationsStore",
                    "store_backend": {
                        "class_name": "GXCloudStoreBackend",
                        "ge_cloud_base_url": r"${GX_CLOUD_BASE_URL}",
                        "ge_cloud_credentials": {
                            "access_token": r"${GX_CLOUD_ACCESS_TOKEN}",
                            "organization_id": r"${GX_CLOUD_ORGANIZATION_ID}",
                        },
                        "ge_cloud_resource_type": "expectation_suite",
                        "suppress_store_backend_id": True,
                    },
                },
                "default_checkpoint_store": {
                    "class_name": "CheckpointStore",
                    "store_backend": {
                        "class_name": "GXCloudStoreBackend",
                        "ge_cloud_base_url": r"${GX_CLOUD_BASE_URL}",
                        "ge_cloud_credentials": {
                            "access_token": r"${GX_CLOUD_ACCESS_TOKEN}",
                            "organization_id": r"${GX_CLOUD_ORGANIZATION_ID}",
                        },
                        "ge_cloud_resource_type": "checkpoint",
                        "suppress_store_backend_id": True,
                    },
                },
                "default_validation_results_store": {
                    "class_name": "ValidationResultsStore",
                    "store_backend": {
                        "class_name": "GXCloudStoreBackend",
                        "ge_cloud_base_url": r"${GX_CLOUD_BASE_URL}",
                        "ge_cloud_credentials": {
                            "access_token": r"${GX_CLOUD_ACCESS_TOKEN}",
                            "organization_id": r"${GX_CLOUD_ORGANIZATION_ID}",
                        },
                        "ge_cloud_resource_type": "validation_result",
                        "suppress_store_backend_id": True,
                    },
                },
            },
        },
        "VALIDATION_DEFINITION_NAMES": set(),
        "validation_definitions": {},
    }


# ##################################
# Cloud API Mock Callbacks
# ##################################

# WARNING: this dict should always be cleared during test teardown
_CLOUD_API_FAKE_DB: FakeDBTypedDict = {}  # type: ignore[typeddict-item] # will be assigned in `create_fake_db_seed_data`


def get_user_id(request: PreparedRequest) -> CallbackResult:
    if not request.url:
        raise NotImplementedError("request.url should not be empty")
    LOGGER.debug(f"{request.method} {request.url}")
    resource_path: str = request.url.split("/")[-1]
    user_dict = _CLOUD_API_FAKE_DB.get(resource_path)
    return CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(user_dict))


def get_dc_configuration_cb(
    request: PreparedRequest,
) -> CallbackResult:
    if not request.url:
        raise NotImplementedError("request.url should not be empty")
    LOGGER.debug(f"{request.method} {request.url}")
    resource_path: str = request.url.split("/")[-1]
    dc_config = _CLOUD_API_FAKE_DB.get(resource_path)
    LOGGER.debug(f"GET response body -->\n{pf(dc_config, depth=2)}")
    if not dc_config:
        raise NotImplementedError(f"{resource_path} should never be empty")
    return CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(dc_config))


def get_datasource_by_id_cb(request: PreparedRequest) -> CallbackResult:
    if not request.url:
        raise NotImplementedError("request.url should not be empty")
    LOGGER.debug(f"{request.method} {request.url}")

    parsed_url = urllib.parse.urlparse(request.url)
    datasource_id = parsed_url.path.split("/")[-1]

    datasource: dict | None = _CLOUD_API_FAKE_DB["datasources"].get(datasource_id)
    if datasource:
        result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(datasource))
    else:
        result = CallbackResult(
            404,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "Mock 404",
                        "detail": f"Datasource {datasource_id} not found",
                        "source": None,
                    }
                ]
            ).json(),
        )
    return result


def delete_datasources_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    datasource_id: str = parsed_url.path.split("/")[-1]  # type: ignore[arg-type,assignment] # FIXME CoP

    datasources: dict[str, dict] = _CLOUD_API_FAKE_DB["datasources"]
    deleted_ds = datasources.pop(datasource_id, None)
    if deleted_ds:
        ds_name = deleted_ds["data"]["name"]
        _CLOUD_API_FAKE_DB["DATASOURCE_NAMES"].remove(ds_name)
        LOGGER.debug(f"Deleted datasource '{ds_name}'")
        result = CallbackResult(204, headers={}, body="")
    else:
        errors = ErrorPayloadSchema(errors=[{"code": "mock 404", "detail": None, "source": None}])
        result = CallbackResult(404, headers=DEFAULT_HEADERS, body=errors.json())
    return result


def delete_data_assets_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    data_asset_id: str = parsed_url.path.split("/")[-1]  # type: ignore[arg-type,assignment] # FIXME CoP

    datasources: dict[str, dict] = _CLOUD_API_FAKE_DB["datasources"]
    deleted_asset_idx = None
    deleted_asset = None

    # find and remove asset from datasource config
    for datasource in datasources.values():
        for idx, asset in enumerate(datasource["data"].get("assets", {})):
            if asset.get("id") == data_asset_id:
                deleted_asset_idx = idx
                break
        if deleted_asset_idx is not None:
            deleted_asset = datasource["data"]["assets"].pop(deleted_asset_idx)
            break

    if deleted_asset:
        asset_name = deleted_asset["name"]
        LOGGER.debug(f"Deleted asset '{asset_name}'")
        result = CallbackResult(204, headers={}, body="")
    else:
        errors = ErrorPayloadSchema(errors=[{"code": "mock 404", "detail": None, "source": None}])
        result = CallbackResult(404, headers=DEFAULT_HEADERS, body=errors.json())
    return result


def post_datasources_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    ds_names: set[str] = _CLOUD_API_FAKE_DB["DATASOURCE_NAMES"]

    if not request.body:
        return CallbackResult(
            400,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[{"code": "400", "detail": "Missing Body", "source": None}]
            ).json(),
        )

    try:
        LOGGER.debug(f"POST request body -->\n{pf(json.loads(request.body), depth=4)}")
        payload = CloudResponseSchema.from_datasource_json(request.body)

        datasource_name: str = payload.data.name
        if datasource_name not in ds_names:
            datasource_id = payload.data.id
            if not datasource_id:
                datasource_id = str(uuid.uuid4())
                payload.data.id = datasource_id
            assert datasource_id not in _CLOUD_API_FAKE_DB["datasources"], (
                f"ID collision for '{datasource_name}'"
            )

            _CLOUD_API_FAKE_DB["datasources"][datasource_id] = payload.dict()
            _CLOUD_API_FAKE_DB["DATASOURCE_NAMES"].add(payload.data.name)

            result = CallbackResult(201, headers=DEFAULT_HEADERS, body=payload.json())
        else:
            errors = ErrorPayloadSchema(
                errors=[
                    {
                        "code": "mock 400/409",
                        "detail": f"Datasource with name '{datasource_name}' already exists.",
                        "source": None,
                    }
                ]
            )
            result = CallbackResult(409, headers=DEFAULT_HEADERS, body=errors.json())

        return result
    except pydantic.ValidationError as val_err:
        LOGGER.exception(val_err)  # noqa: TRY401 # FIXME CoP
        return CallbackResult(
            400,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "mock 400",
                        "detail": str(val_err.errors()),
                        "source": None,
                    }
                ]
            ).json(),
        )
    except Exception as err:
        return CallbackResult(
            500,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[{"code": "mock 500", "detail": repr(err), "source": None}]
            ).json(),
        )


def put_datasource_cb(request: PreparedRequest) -> CallbackResult:  # noqa: C901 # FIXME CoP
    LOGGER.debug(f"{request.method} {request.url}")
    if not request.url:
        raise NotImplementedError("request.url should not be empty")

    if not request.body:
        errors = ErrorPayloadSchema(
            errors=[{"code": "mock 400", "detail": "missing body", "source": None}]
        )
        return CallbackResult(400, headers=DEFAULT_HEADERS, body=errors.json())

    LOGGER.debug(f"PUT request body -->\n{pf(json.loads(request.body), depth=4)}")
    payload = CloudResponseSchema.from_datasource_json(request.body)

    parsed_url = urllib.parse.urlparse(request.url)
    datasource_id = parsed_url.path.split("/")[-1]

    # Ensure that assets and batch definitions get IDs
    # Note that this logic should also happen in POST but is not implemented for our fake
    for asset in payload.data.assets:
        if not asset.get("id"):
            asset["id"] = str(uuid.uuid4())
        for batch_definition in asset.get("batch_definitions", []):
            if not batch_definition.get("id"):
                batch_definition["id"] = str(uuid.uuid4())

    old_datasource: dict | None = _CLOUD_API_FAKE_DB["datasources"].get(datasource_id)
    if old_datasource:
        if payload.data.name != old_datasource["data"]["name"]:
            raise NotImplementedError("Unsure how to handle name change")
        _CLOUD_API_FAKE_DB["datasources"][datasource_id] = payload.dict()
        result = CallbackResult(200, headers=DEFAULT_HEADERS, body=payload.json())
    else:
        result = CallbackResult(404, headers=DEFAULT_HEADERS, body="")
    return result


def get_datasources_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    query_params = urllib.parse.parse_qs(parsed_url.query)  # type: ignore[type-var] # FIXME CoP
    queried_names: Sequence[str] = query_params.get("name", [])  # type: ignore[assignment] # FIXME CoP

    all_datasources: dict[str, dict] = _CLOUD_API_FAKE_DB["datasources"]
    datasources_list: list[dict] = list(all_datasources.values())
    if queried_names:
        datasources_list = [
            d["data"] for d in datasources_list if d["data"]["name"] in queried_names
        ]
    else:
        datasources_list = [d["data"] for d in datasources_list]

    resp_body = {"data": datasources_list}
    result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(resp_body))
    LOGGER.debug(f"Response {result.status}")
    return result


def get_expectation_suites_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    query_params = urllib.parse.parse_qs(parsed_url.query)  # type: ignore[type-var] # FIXME CoP
    queried_names: Sequence[str] = query_params.get("name", [])  # type: ignore[assignment] # FIXME CoP

    exp_suites: dict[str, dict] = _CLOUD_API_FAKE_DB["expectation_suites"]
    exp_suite_list: list[dict] = [d["data"] for d in exp_suites.values()]
    if queried_names:
        exp_suite_list = [d for d in exp_suite_list if d["name"] in queried_names]

    resp_body = {"data": exp_suite_list}

    result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(resp_body))
    LOGGER.debug(f"Response {result.status}")
    return result


def get_expectation_suite_by_id_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    expectation_id: str = parsed_url.path.split("/")[-1]  # type: ignore[arg-type,assignment] # FIXME CoP

    expectation_suite: dict | None = _CLOUD_API_FAKE_DB["expectation_suites"].get(expectation_id)
    if expectation_suite:
        result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(expectation_suite))
    else:
        result = CallbackResult(404, headers=DEFAULT_HEADERS, body="")
    return result


def post_expectation_suites_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    if not request.body:
        raise NotImplementedError("Handling missing body")

    payload: dict = json.loads(request.body)
    name = payload["data"]["name"]

    exp_suite_names: set[str] = _CLOUD_API_FAKE_DB["EXPECTATION_SUITE_NAMES"]
    exp_suites: dict[str, dict] = _CLOUD_API_FAKE_DB["expectation_suites"]

    if name in exp_suite_names:
        result = CallbackResult(
            409,  # not really a 409 in prod but it's a more informative status code
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "mock 400/409",
                        "detail": f"'{name}' already defined",
                        "source": None,
                    }
                ]
            ).json(),
        )
    else:
        suite_id = str(uuid.uuid4())
        payload["data"]["id"] = suite_id
        payload["data"]["id"] = suite_id
        for expectation_configuration in payload["data"]["expectations"]:
            expectation_configuration["id"] = str(uuid.uuid4())
        exp_suites[suite_id] = payload
        exp_suite_names.add(name)
        result = CallbackResult(201, headers=DEFAULT_HEADERS, body=json.dumps(payload))

    LOGGER.debug(f"Response {result.status}")
    return result


def put_expectation_suites_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    if not request.body:
        errors = ErrorPayloadSchema(
            errors=[{"code": "mock 400", "detail": "missing body", "source": None}]
        )
        return CallbackResult(400, headers=DEFAULT_HEADERS, body=errors.json())

    payload: dict = json.loads(request.body)
    parsed_url = urllib.parse.urlparse(request.url)
    suite_id: str = parsed_url.path.split("/")[-1]  # type: ignore[arg-type,assignment] # FIXME CoP

    name = payload["data"]["name"]

    exp_suite_names: set[str] = _CLOUD_API_FAKE_DB["EXPECTATION_SUITE_NAMES"]
    exp_suites: dict[str, dict] = _CLOUD_API_FAKE_DB["expectation_suites"]

    old_suite = exp_suites.get(suite_id)

    if not old_suite:
        result = CallbackResult(
            404,  # not really a 409 in prod but it's a more informative status code
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "404",
                        "detail": "Suite not found",
                        "source": None,
                    }
                ]
            ).json(),
        )
    else:
        payload["data"]["id"] = suite_id
        payload["data"]["id"] = suite_id
        for expectation_configuration in payload["data"]["expectations"]:
            # add IDs to new expectations
            if not expectation_configuration.get("id"):
                expectation_configuration["id"] = str(uuid.uuid4())
        exp_suites[suite_id] = payload
        exp_suite_names.add(name)
        result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(payload))

    LOGGER.debug(f"Response {result.status}")
    return result


def delete_expectation_suites_cb(request: PreparedRequest) -> CallbackResult:
    parsed_url = urllib.parse.urlparse(request.url)
    suite_id: str = parsed_url.path.split("/")[-1]  # type: ignore[arg-type,assignment] # FIXME CoP
    exp_suites: dict[str, dict] = _CLOUD_API_FAKE_DB["expectation_suites"]
    old_suite = exp_suites.pop(suite_id, None)
    if not old_suite:
        result = CallbackResult(
            404,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "404",
                        "detail": "Suite not found",
                        "source": None,
                    }
                ]
            ).json(),
        )
    else:
        result = CallbackResult(204, headers={}, body="")
    return result


def get_checkpoints_cb(requests: PreparedRequest) -> CallbackResult:
    url = requests.url
    LOGGER.debug(f"{requests.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    query_params = urllib.parse.parse_qs(parsed_url.query)  # type: ignore[type-var] # FIXME CoP
    queried_names: Sequence[str] = query_params.get("name", [])  # type: ignore[assignment] # FIXME CoP

    checkpoints: dict[str, dict] = _CLOUD_API_FAKE_DB["checkpoints"]
    checkpoint_list: list[dict] = list(checkpoints.values())
    if queried_names:
        checkpoint_list = [d for d in checkpoint_list if d["name"] in queried_names]

    resp_body = {"data": checkpoint_list}

    result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(resp_body))
    LOGGER.debug(f"Response {result.status}")
    return result


def get_checkpoint_by_id_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    checkpoint_id: str = parsed_url.path.split("/")[-1]  # type: ignore[arg-type,assignment] # FIXME CoP

    if checkpoint := _CLOUD_API_FAKE_DB["checkpoints"].get(checkpoint_id):
        resp_body = json.dumps({"data": checkpoint})
        result = CallbackResult(200, headers=DEFAULT_HEADERS, body=resp_body)
    else:
        result = CallbackResult(
            404,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "mock 404",
                        "detail": f"Checkpoint '{checkpoint_id}' not found",
                        "source": None,
                    }
                ]
            ).json(),
        )
    return result


def post_checkpoints_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    if not request.body:
        raise NotImplementedError("Handling missing body")

    payload: dict = json.loads(request.body)
    name = payload["data"]["name"]

    checkpoints: dict[str, dict] = _CLOUD_API_FAKE_DB["checkpoints"]
    checkpoint_names: set[str] = _CLOUD_API_FAKE_DB["CHECKPOINT_NAMES"]

    if name in checkpoint_names:
        result = CallbackResult(
            409,  # not really a 409 in prod but it's a more informative status code
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "mock 400/409",
                        "detail": f"'{name}' already defined",
                        "source": None,
                    }
                ]
            ).json(),
        )
    else:
        id_ = str(uuid.uuid4())
        payload["data"]["id"] = id_
        checkpoints[id_] = payload["data"]
        checkpoint_names.add(name)
        result = CallbackResult(201, headers=DEFAULT_HEADERS, body=json.dumps(payload))

    LOGGER.debug(f"Response {result.status}")
    return result


def put_checkpoint_cb(request: PreparedRequest) -> CallbackResult:
    if not request.body:
        raise NotImplementedError("Handling missing body")

    payload: dict = json.loads(request.body)
    name = payload["data"]["name"]

    checkpoints: dict[str, dict] = _CLOUD_API_FAKE_DB["checkpoints"]
    checkpoint_names: set[str] = _CLOUD_API_FAKE_DB["CHECKPOINT_NAMES"]

    if name not in checkpoint_names:
        result = CallbackResult(
            404,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "404",
                        "detail": f"Checkpoint '{name}' not found",
                        "source": None,
                    }
                ]
            ).json(),
        )
    else:
        checkpoint_id = payload["data"]["id"]
        checkpoints[checkpoint_id] = payload["data"]
        result = CallbackResult(200, headers=DEFAULT_HEADERS, body=json.dumps(payload))
    return result


def delete_checkpoint_by_id_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    path = str(parsed_url.path)
    checkpoint_id = path.split("/")[-1]
    checkpoints: dict[str, dict] = _CLOUD_API_FAKE_DB["checkpoints"]

    deleted_cp = checkpoints.pop(checkpoint_id, None)
    if not deleted_cp:
        errors = ErrorPayloadSchema(errors=[{"code": "mock 404", "detail": None, "source": None}])
        return CallbackResult(404, headers=DEFAULT_HEADERS, body=errors.json())

    print(pf(deleted_cp, depth=5))
    cp_name = deleted_cp["name"]
    _CLOUD_API_FAKE_DB["CHECKPOINT_NAMES"].remove(cp_name)
    LOGGER.debug(f"Deleted checkpoint '{cp_name}'")
    return CallbackResult(204, headers={}, body="")


def delete_checkpoint_by_name_cb(
    request: PreparedRequest,
) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    parsed_url = urllib.parse.urlparse(url)
    query_params = urllib.parse.parse_qs(parsed_url.query)  # type: ignore[type-var] # FIXME CoP
    queried_names: list[str] = query_params.get("name", [])  # type: ignore[assignment] # FIXME CoP
    if not queried_names:
        raise ValueError("Must provide checkpoint name for deletion.")

    cp_name = queried_names[0]
    checkpoints: dict[str, dict] = _CLOUD_API_FAKE_DB["checkpoints"]

    checkpoint_id: str | None = None
    for checkpoint in checkpoints.values():
        if checkpoint["name"] == cp_name:
            checkpoint_id = checkpoint["id"]
            break

    if not checkpoint_id:
        errors = ErrorPayloadSchema(errors=[{"code": "mock 404", "detail": None, "source": None}])
        return CallbackResult(404, headers=DEFAULT_HEADERS, body=errors.json())

    deleted_cp = checkpoints.pop(checkpoint_id)
    print(pf(deleted_cp, depth=5))

    _CLOUD_API_FAKE_DB["CHECKPOINT_NAMES"].remove(cp_name)
    LOGGER.debug(f"Deleted checkpoint '{cp_name}'")
    return CallbackResult(204, headers={}, body="")


def post_validation_results_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    if not request.body:
        raise NotImplementedError("Handling missing body")

    payload: dict = json.loads(request.body)
    validation_id = payload["data"]["meta"]["validation_id"]
    if validation_id:
        return CallbackResult(
            200,
            headers=DEFAULT_HEADERS,
            body=json.dumps(
                {
                    "data": {
                        "id": validation_id,
                        "result_url": (
                            f"{GX_CLOUD_MOCK_BASE_URL}organizations/{FAKE_ORG_ID}/"
                            "validation-results/{validation_id}"
                        ),
                    }
                }
            ),
        )
    else:
        result = CallbackResult(
            422,  # 400 in prod but this is a more informative status code
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "Mock 400/422",
                        "detail": "Field may not be null.",
                        "source": {"pointer": "/data/attributes/result/meta/validation_id"},
                    }
                ]
            ).json(),
        )

    LOGGER.debug(f"Response {result.status}")
    return result


def get_validation_definition_by_id_cb(request: PreparedRequest) -> CallbackResult:
    if not request.url:
        raise NotImplementedError("request.url should not be empty")
    LOGGER.debug(f"{request.method} {request.url}")

    parsed_url = urllib.parse.urlparse(request.url)
    validation_id = parsed_url.path.split("/")[-1]

    validation_definition: dict | None = _CLOUD_API_FAKE_DB["validation_definitions"].get(
        validation_id
    )
    if validation_definition:
        validation_definition = {"data": validation_definition}
        result = CallbackResult(
            200, headers=DEFAULT_HEADERS, body=json.dumps(validation_definition)
        )
    else:
        result = CallbackResult(
            404,
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "Mock 404",
                        "detail": f"ValidationDefinition {validation_id} not found",
                        "source": None,
                    }
                ]
            ).json(),
        )
    return result


def post_validation_definitions_cb(request: PreparedRequest) -> CallbackResult:
    url = request.url
    LOGGER.debug(f"{request.method} {url}")

    if not request.body:
        raise NotImplementedError("Handling missing body")

    payload: dict = json.loads(request.body)
    name = payload["data"]["name"]

    validation_definitions: dict[str, dict] = _CLOUD_API_FAKE_DB["validation_definitions"]
    validation_definition_names: set[str] = _CLOUD_API_FAKE_DB["VALIDATION_DEFINITION_NAMES"]

    if name in validation_definitions:
        result = CallbackResult(
            409,  # not really a 409 in prod but it's a more informative status code
            headers=DEFAULT_HEADERS,
            body=ErrorPayloadSchema(
                errors=[
                    {
                        "code": "mock 400/409",
                        "detail": f"'{name}' already defined",
                        "source": None,
                    }
                ]
            ).json(),
        )
    else:
        id_ = str(uuid.uuid4())
        payload["data"]["id"] = id_
        validation_definitions[id_] = payload["data"]
        validation_definition_names.add(name)
        result = CallbackResult(201, headers=DEFAULT_HEADERS, body=json.dumps(payload))

    LOGGER.debug(f"Response {result.status}")
    return result


# #######################
# RequestMocks
# #######################


@contextlib.contextmanager
def gx_cloud_api_fake_ctx(
    cloud_details: CloudDetails,
    fds_config: GxConfig | None = None,
    assert_all_requests_are_fired: bool = False,
) -> Generator[responses.RequestsMock, None, None]:
    """Mock the GX Cloud API for the lifetime of the context manager."""
    org_url_base_V0 = urllib.parse.urljoin(
        cloud_details.base_url, f"organizations/{cloud_details.org_id}/"
    )
    org_url_base_V1 = urllib.parse.urljoin(
        cloud_details.base_url, f"api/v1/organizations/{cloud_details.org_id}/"
    )
    dc_config_url = urllib.parse.urljoin(org_url_base_V1, "data-context-configuration")
    me_url = urllib.parse.urljoin(org_url_base_V0, "accounts/me")

    assert not _CLOUD_API_FAKE_DB, "_CLOUD_API_FAKE_DB should be empty"
    _CLOUD_API_FAKE_DB.update(create_fake_db_seed_data(fds_config))

    LOGGER.info("Mocking the GX Cloud API")

    with responses.RequestsMock(
        assert_all_requests_are_fired=assert_all_requests_are_fired
    ) as resp_mocker:
        resp_mocker.add_callback(responses.GET, me_url, get_user_id)
        resp_mocker.add_callback(responses.GET, dc_config_url, get_dc_configuration_cb)
        resp_mocker.add_callback(
            responses.GET,
            urllib.parse.urljoin(org_url_base_V1, "datasources"),
            get_datasources_cb,
        )
        resp_mocker.add_callback(
            responses.POST,
            urllib.parse.urljoin(org_url_base_V1, "datasources"),
            post_datasources_cb,
        )
        resp_mocker.add_callback(
            responses.GET,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"datasources/{UUID_REGEX}")),
            get_datasource_by_id_cb,
        )
        resp_mocker.add_callback(
            responses.DELETE,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"datasources/{UUID_REGEX}")),
            delete_datasources_cb,
        )
        resp_mocker.add_callback(
            responses.PUT,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"datasources/{UUID_REGEX}")),
            put_datasource_cb,
        )
        resp_mocker.add_callback(
            responses.DELETE,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"data-assets/{UUID_REGEX}")),
            delete_data_assets_cb,
        )
        resp_mocker.add_callback(
            responses.GET,
            urllib.parse.urljoin(org_url_base_V1, "expectation-suites"),
            get_expectation_suites_cb,
        )
        resp_mocker.add_callback(
            responses.GET,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"expectation-suites/{UUID_REGEX}")),
            get_expectation_suite_by_id_cb,
        )
        resp_mocker.add_callback(
            responses.POST,
            urllib.parse.urljoin(org_url_base_V1, "expectation-suites"),
            post_expectation_suites_cb,
        )
        resp_mocker.add_callback(
            responses.PUT,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"expectation-suites/{UUID_REGEX}")),
            put_expectation_suites_cb,
        )
        resp_mocker.add_callback(
            responses.DELETE,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"expectation-suites/{UUID_REGEX}")),
            delete_expectation_suites_cb,
        )
        resp_mocker.add_callback(
            responses.GET,
            urllib.parse.urljoin(org_url_base_V1, "checkpoints"),
            get_checkpoints_cb,
        )
        resp_mocker.add_callback(
            responses.POST,
            urllib.parse.urljoin(org_url_base_V1, "checkpoints"),
            post_checkpoints_cb,
        )
        resp_mocker.add_callback(
            responses.DELETE,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"checkpoints/{UUID_REGEX}")),
            delete_checkpoint_by_id_cb,
        )
        resp_mocker.add_callback(
            responses.DELETE,
            urllib.parse.urljoin(org_url_base_V1, "checkpoints"),
            delete_checkpoint_by_name_cb,
        )
        resp_mocker.add_callback(
            responses.GET,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"checkpoints/{UUID_REGEX}")),
            get_checkpoint_by_id_cb,
        )
        resp_mocker.add_callback(
            responses.PUT,
            re.compile(urllib.parse.urljoin(org_url_base_V1, f"checkpoints/{UUID_REGEX}")),
            put_checkpoint_cb,
        )
        resp_mocker.add_callback(
            responses.POST,
            urllib.parse.urljoin(org_url_base_V1, "validation-results"),
            post_validation_results_cb,
        )
        resp_mocker.add_callback(
            responses.POST,
            urllib.parse.urljoin(org_url_base_V1, "validation-definitions"),
            post_validation_definitions_cb,
        )
        resp_mocker.add_callback(
            responses.GET,
            re.compile(
                urllib.parse.urljoin(org_url_base_V1, f"validation-definitions/{UUID_REGEX}")
            ),
            get_validation_definition_by_id_cb,
        )

        yield resp_mocker

    LOGGER.info(f"GX Cloud API Mock ending state ->\n{pf(_CLOUD_API_FAKE_DB, depth=2)}")
    _CLOUD_API_FAKE_DB.clear()



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Mapping, Optional

import pytest

from great_expectations.compatibility.pydantic import BaseSettings
from great_expectations.compatibility.typing_extensions import override
from great_expectations.data_context import AbstractDataContext
from great_expectations.datasource.fluent.sql_datasource import TableAsset
from tests.integration.sql_session_manager import SessionSQLEngineManager
from tests.integration.test_utils.data_source_config.base import (
    BatchTestSetup,
    DataSourceTestConfig,
)
from tests.integration.test_utils.data_source_config.sql import SQLBatchTestSetup

if TYPE_CHECKING:
    import pandas as pd


class BigQueryDatasourceTestConfig(DataSourceTestConfig):
    @property
    @override
    def label(self) -> str:
        return "big-query"

    @property
    @override
    def pytest_mark(self) -> pytest.MarkDecorator:
        return pytest.mark.bigquery

    @override
    def create_batch_setup(
        self,
        request: pytest.FixtureRequest,
        data: pd.DataFrame,
        extra_data: Mapping[str, pd.DataFrame],
        context: AbstractDataContext,
        engine_manager: Optional[SessionSQLEngineManager] = None,
    ) -> BatchTestSetup:
        return BigQueryBatchTestSetup(
            data=data,
            config=self,
            extra_data=extra_data,
            table_name=self.table_name,
            context=context,
            engine_manager=engine_manager,
        )


class BigQueryBatchTestSetup(SQLBatchTestSetup[BigQueryDatasourceTestConfig]):
    @property
    @override
    def connection_string(self) -> str:
        return self.big_query_connection_config.connection_string

    @property
    @override
    def use_schema(self) -> bool:
        # BigQuery calls its schemas "datasets". Their docs show that the sql way of defining a
        # dataset is to create a schema: https://cloud.google.com/bigquery/docs/datasets#sql
        return True

    @override
    def make_asset(self) -> TableAsset:
        return self.context.data_sources.add_bigquery(
            name=self._random_resource_name(), connection_string=self.connection_string
        ).add_table_asset(
            name=self._random_resource_name(),
            table_name=self.table_name,
            schema_name=self.schema,
        )

    @cached_property
    def big_query_connection_config(self) -> BigQueryConnectionConfig:
        return BigQueryConnectionConfig()  # type: ignore[call-arg]  # retrieves env vars


class BigQueryConnectionConfig(BaseSettings):
    """Environment variables for BigQuery connection.
    These are injected in via CI, but when running locally, you may use your own credentials.
    GOOGLE_APPLICATION_CREDENTIALS must be kept secret
    """

    GE_TEST_GCP_PROJECT: str
    GE_TEST_BIGQUERY_DATASET: str
    GOOGLE_APPLICATION_CREDENTIALS: str

    @property
    def connection_string(self) -> str:
        return f"bigquery://{self.GE_TEST_GCP_PROJECT}/{self.GE_TEST_BIGQUERY_DATASET}?credentials_path={self.GOOGLE_APPLICATION_CREDENTIALS}"



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

import pandas as pd

from great_expectations import ValidationDefinition
from great_expectations.checkpoint import Checkpoint
from great_expectations.core import ExpectationSuite
from great_expectations.expectations import ExpectColumnValuesToNotBeNull
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config.snowflake import SnowflakeDatasourceTestConfig

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult


@parameterize_batch_for_data_sources(
    data_source_configs=[SnowflakeDatasourceTestConfig()],
    data=pd.DataFrame({"test_e2e_sf": [1, 2]}),
)
def test_checkpoint_run(batch_for_datasource):
    """Test running a Checkpoint end-to-end"""
    context = batch_for_datasource.datasource.data_context
    expectation_suite = context.suites.add(
        ExpectationSuite(
            name=f"es_{uuid.uuid4().hex}",
            expectations=[ExpectColumnValuesToNotBeNull(column="test_e2e_sf", mostly=1)],
        )
    )
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(
            name=f"val_def_{uuid.uuid4().hex}",
            data=batch_for_datasource.data_asset.batch_definitions[0],
            suite=expectation_suite,
        )
    )
    checkpoint = context.checkpoints.add(
        Checkpoint(
            name=f"E2E Test Checkpoint {uuid.uuid4().hex}",
            validation_definitions=[validation_definition],
        )
    )
    checkpoint_result: CheckpointResult = checkpoint.run()
    assert checkpoint_result.success



from __future__ import annotations

import logging
import os
import pathlib
import re
import tempfile
from mimetypes import guess_type
from pathlib import Path
from typing import Optional
from zipfile import ZipFile, is_zipfile

from great_expectations.core.data_context_key import DataContextKey
from great_expectations.data_context.store.gx_cloud_store_backend import (
    GXCloudStoreBackend,
)
from great_expectations.data_context.store.tuple_store_backend import TupleStoreBackend
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    SiteSectionIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.data_context.util import (
    file_relative_path,
    instantiate_class_from_config,
    load_class,
)
from great_expectations.exceptions import ClassInstantiationError, DataContextError
from great_expectations.util import (
    filter_properties_dict,
    verify_dynamic_loading_support,
)

logger = logging.getLogger(__name__)


class HtmlSiteStore:
    """
    A HtmlSiteStore facilitates publishing rendered documentation built from Expectation Suites, Profiling Results, and Validation Results.

    --ge-feature-maturity-info--

        id: html_site_store_filesystem
        title: HTML Site Store - Filesystem
        icon:
        short_description: DataDocs on Filesystem
        description: For publishing rendered documentation built from Expectation Suites, Profiling Results, and Validation Results on the Filesystem
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/configuring_data_docs/how_to_host_and_share_data_docs_on_a_filesystem.html
        maturity: Production
        maturity_details:
            api_stability: Mostly Stable (profiling)
            implementation_completeness: Complete
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: N/A
            documentation_completeness: Partial
            bug_risk: Low

        id: html_site_store_s3
        title: HTML Site Store - S3
        icon:
        short_description: DataDocs on S3
        description: For publishing rendered documentation built from Expectation Suites, Profiling Results, and Validation Results on S3
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/configuring_data_docs/how_to_host_and_share_data_docs_on_s3.html
        maturity: Beta
        maturity_details:
            api_stability: Mostly Stable (profiling)
            implementation_completeness: Complete
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: Minimal
            documentation_completeness: Complete
            bug_risk: Moderate

        id: html_site_store_gcs
        title: HTMLSiteStore - GCS
        icon:
        short_description: DataDocs on GCS
        description: For publishing rendered documentation built from Expectation Suites, Profiling Results, and Validation Results on GCS
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/configuring_data_docs/how_to_host_and_share_data_docs_on_gcs.html
        maturity: Beta
        maturity_details:
            api_stability: Mostly Stable (profiling)
            implementation_completeness: Complete
            unit_test_coverage: Complete
            integration_infrastructure_test_coverage: Minimal
            documentation_completeness: Partial (needs auth)
            bug_risk: Moderate (resource URL may have bugs)

        id: html_site_store_azure_blob_storage
        title: HTMLSiteStore - Azure
        icon:
        short_description: DataDocs on Azure Blob Storage
        description: For publishing rendered documentation built from Expectation Suites, Profiling Results, and Validation Results on Azure Blob Storage
        how_to_guide_url: https://docs.greatexpectations.io/en/latest/how_to_guides/configuring_data_docs/how_to_host_and_share_data_docs_on_azure_blob_storage.html
        maturity: N/A
        maturity_details:
            api_stability: Mostly Stable (profiling)
            implementation_completeness: Minimal
            unit_test_coverage: Minimal
            integration_infrastructure_test_coverage: Minimal
            documentation_completeness: Minimal
            bug_risk: Moderate

    --ge-feature-maturity-info--
    """  # noqa: E501 # FIXME CoP

    _key_class = SiteSectionIdentifier

    def __init__(  # noqa: C901 #  11
        self, store_backend=None, runtime_environment=None
    ) -> None:
        store_backend_module_name = store_backend.get(
            "module_name", "great_expectations.data_context.store"
        )
        store_backend_class_name = store_backend.get("class_name", "TupleFilesystemStoreBackend")
        verify_dynamic_loading_support(module_name=store_backend_module_name)
        store_class = load_class(store_backend_class_name, store_backend_module_name)

        # Store Class was loaded successfully; verify that it is of a correct subclass.
        if not issubclass(store_class, (TupleStoreBackend, GXCloudStoreBackend)):
            raise DataContextError(  # noqa: TRY003 # FIXME CoP
                f"Invalid configuration: HtmlSiteStore needs a {TupleStoreBackend.__name__} or {GXCloudStoreBackend.__name__}"  # noqa: E501 # FIXME CoP
            )
        if "filepath_template" in store_backend or (
            "fixed_length_key" in store_backend and store_backend["fixed_length_key"] is True
        ):
            logger.warning(
                "Configuring a filepath_template or using fixed_length_key is not supported in SiteBuilder: "  # noqa: E501 # FIXME CoP
                "filepaths will be selected based on the type of asset rendered."
            )

        # One thing to watch for is reversibility of keys.
        # If several types are being written to overlapping directories, we could get collisions.
        module_name = "great_expectations.data_context.store"
        filepath_suffix = ".html"
        is_gx_cloud_store = store_backend["class_name"] == GXCloudStoreBackend.__name__
        expectation_config_defaults = {
            "module_name": module_name,
            "filepath_prefix": "expectations",
            "filepath_suffix": filepath_suffix,
            "suppress_store_backend_id": True,
        }
        if is_gx_cloud_store:
            expectation_config_defaults = {
                "module_name": module_name,
                "suppress_store_backend_id": True,
            }
        expectation_suite_identifier_obj = instantiate_class_from_config(
            config=store_backend,
            runtime_environment=runtime_environment,
            config_defaults=expectation_config_defaults,
        )
        if not expectation_suite_identifier_obj:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=store_backend["class_name"],
            )

        validation_result_config_defaults = {
            "module_name": module_name,
            "filepath_prefix": "validations",
            "filepath_suffix": filepath_suffix,
            "suppress_store_backend_id": True,
        }
        if is_gx_cloud_store:
            validation_result_config_defaults = {
                "module_name": module_name,
                "suppress_store_backend_id": True,
            }

        validation_result_idendifier_obj = instantiate_class_from_config(
            config=store_backend,
            runtime_environment=runtime_environment,
            config_defaults=validation_result_config_defaults,
        )
        if not validation_result_idendifier_obj:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=store_backend["class_name"],
            )

        filepath_template = "index.html"
        index_page_config_defaults = {
            "module_name": module_name,
            "filepath_template": filepath_template,
            "suppress_store_backend_id": True,
        }
        if is_gx_cloud_store:
            index_page_config_defaults = {
                "module_name": module_name,
                "suppress_store_backend_id": True,
            }

        index_page_obj = instantiate_class_from_config(
            config=store_backend,
            runtime_environment=runtime_environment,
            config_defaults=index_page_config_defaults,
        )
        if not index_page_obj:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=store_backend["class_name"],
            )

        static_assets_config_defaults = {
            "module_name": module_name,
            "filepath_template": None,
            "suppress_store_backend_id": True,
        }
        if is_gx_cloud_store:
            static_assets_config_defaults = {
                "module_name": module_name,
                "suppress_store_backend_id": True,
            }
        static_assets_obj = instantiate_class_from_config(
            config=store_backend,
            runtime_environment=runtime_environment,
            config_defaults=static_assets_config_defaults,
        )
        if not static_assets_obj:
            raise ClassInstantiationError(
                module_name=module_name,
                package_name=None,
                class_name=store_backend["class_name"],
            )

        self.store_backends = {
            ExpectationSuiteIdentifier: expectation_suite_identifier_obj,
            ValidationResultIdentifier: validation_result_idendifier_obj,
            "index_page": index_page_obj,
            "static_assets": static_assets_obj,
        }

        # NOTE: Instead of using the filesystem as the source of record for keys,
        # this class tracks keys separately in an internal set.
        # This means that keys are stored for a specific session, but can't be fetched after the original  # noqa: E501 # FIXME CoP
        # HtmlSiteStore instance leaves scope.
        # Doing it this way allows us to prevent namespace collisions among keys while still having multiple  # noqa: E501 # FIXME CoP
        # backends that write to the same directory structure.
        # It's a pretty reasonable way for HtmlSiteStore to do its job---you just have to remember that it  # noqa: E501 # FIXME CoP
        # can't necessarily set and list_keys like most other Stores.
        self.keys = set()  # type: ignore[var-annotated] # FIXME CoP

        # Gather the call arguments of the present function (include the "module_name" and add the "class_name"), filter  # noqa: E501 # FIXME CoP
        # out the Falsy values, and set the instance "_config" variable equal to the resulting dictionary.  # noqa: E501 # FIXME CoP
        self._config = {
            "store_backend": store_backend,
            "runtime_environment": runtime_environment,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    def get(self, key):
        self._validate_key(key)
        return self.store_backends[type(key.resource_identifier)].get(key.to_tuple())

    def set(self, key, serialized_value):
        self._validate_key(key)
        self.keys.add(key)

        return self.store_backends[type(key.resource_identifier)].set(
            key.resource_identifier.to_tuple(),
            serialized_value,
            content_encoding="utf-8",
            content_type="text/html; charset=utf-8",
        )

    def get_url_for_resource(self, resource_identifier=None, only_if_exists=True) -> Optional[str]:
        """
        Return the URL of the HTML document that renders a resource
        (e.g., an expectation suite or a validation result).

        :param resource_identifier: ExpectationSuiteIdentifier, ValidationResultIdentifier
                or any other type's identifier. The argument is optional - when
                not supplied, the method returns the URL of the index page.
        :return: URL (string)
        """
        if resource_identifier is None:
            store_backend = self.store_backends["index_page"]
            key = ()
        elif isinstance(resource_identifier, ExpectationSuiteIdentifier):
            store_backend = self.store_backends[ExpectationSuiteIdentifier]
            key = resource_identifier.to_tuple()
        elif isinstance(resource_identifier, ValidationResultIdentifier):
            store_backend = self.store_backends[ValidationResultIdentifier]
            key = resource_identifier.to_tuple()
        else:
            # this method does not support getting the URL of static assets
            raise ValueError(f"Cannot get URL for resource {resource_identifier!s:s}")  # noqa: TRY003 # FIXME CoP

        # <WILL> : this is a hack for Taylor. Change this back. 20200924
        # if only_if_exists:
        #    return (
        #        store_backend.get_url_for_key(key)
        #        if store_backend.has_key(key)
        #        else None
        #    )
        # return store_backend.get_url_for_key(key)

        if store_backend.base_public_path:
            if only_if_exists:
                return (
                    store_backend.get_public_url_for_key(key)
                    if store_backend.has_key(key)
                    else None
                )
            else:
                return store_backend.get_public_url_for_key(key)
        else:  # noqa: PLR5501 # FIXME CoP
            if only_if_exists:
                return store_backend.get_url_for_key(key) if store_backend.has_key(key) else None
            else:
                return store_backend.get_url_for_key(key)

    def _validate_key(self, key):
        if not isinstance(key, SiteSectionIdentifier):
            raise TypeError(f"key: {key!r} must be a SiteSectionIdentifier, not {type(key)!r}")  # noqa: TRY003 # FIXME CoP

        for key_class in self.store_backends:
            try:
                if isinstance(key.resource_identifier, key_class):
                    return

            except TypeError:
                # it's ok to have a key that is not a type (e.g. the string "index_page")
                continue

        # The key's resource_identifier didn't match any known key_class
        raise TypeError(  # noqa: TRY003 # FIXME CoP
            f"resource_identifier in key: {key!r} must one of {set(self.store_backends.keys())}, not {type(key)!r}"  # noqa: E501 # FIXME CoP
        )

    def list_keys(self):
        keys = []
        for type_, backend in self.store_backends.items():
            try:
                # If the store_backend does not support list_keys...
                key_tuples = backend.list_keys()
            except NotImplementedError:
                pass
            try:
                if issubclass(type_, DataContextKey):
                    keys += [type_.from_tuple(tuple_) for tuple_ in key_tuples]
            except TypeError:
                # If the key in store_backends is not itself a type...
                pass
        return keys

    def write_index_page(self, page):
        """This third param_store has a special method, which uses a zero-length tuple as a key."""
        return self.store_backends["index_page"].set(
            (),
            page,
            content_encoding="utf-8",
            content_type="text/html; charset=utf-8",
        )

    def clean_site(self) -> None:
        for _, target_store_backend in self.store_backends.items():
            keys = target_store_backend.list_keys()
            for key in keys:
                target_store_backend.remove_key(key)

    def copy_static_assets(  # noqa: C901 #  11
        self, static_assets_source_dir: str | None = None
    ):
        """
        Copies static assets, using a special "static_assets" backend store that accepts variable-length tuples as
        keys, with no filepath_template.
        """  # noqa: E501 # FIXME CoP
        file_exclusions: list[str] = [".DS_Store"]
        dir_exclusions: list[str] = []

        if not static_assets_source_dir:
            static_assets_source_dir = file_relative_path(
                __file__,
                os.path.join("..", "..", "render", "view", "static"),  # noqa: PTH118 # FIXME CoP
            )

        # If `static_assets_source_absdir` contains the string ".zip", then we try to extract (unzip)  # noqa: E501 # FIXME CoP
        # the static files. If the unzipping is successful, that means that Great Expectations is
        # installed into a zip file (see PEP 273) and we need to run this function again
        if ".zip" in static_assets_source_dir.lower():
            unzip_destdir = tempfile.mkdtemp()
            unzipped_ok = self._unzip_assets(static_assets_source_dir, unzip_destdir)
            if unzipped_ok:
                return self.copy_static_assets(unzip_destdir)

        for item in Path(static_assets_source_dir).iterdir():
            item_name = item.name
            # Directory
            if item.is_dir():
                if item_name in dir_exclusions:
                    continue
                # Recurse
                new_source_dir = os.path.join(  # noqa: PTH118 # FIXME CoP
                    static_assets_source_dir, item_name
                )
                self.copy_static_assets(new_source_dir)
            # File
            else:
                # Copy file over using static assets store backend
                if item_name in file_exclusions:
                    continue
                source_name = os.path.join(  # noqa: PTH118 # FIXME CoP
                    static_assets_source_dir, item_name
                )
                with open(source_name, "rb") as f:
                    # Only use path elements starting from static/ for key
                    store_key: tuple[str, ...] = pathlib.Path(source_name).parts
                    store_key = store_key[store_key.index("static") :]
                    content_type, content_encoding = guess_type(item_name, strict=False)

                    if content_type is None:
                        # Use GX-known content-type if possible
                        if source_name.endswith(".otf"):
                            content_type = "font/opentype"
                        else:
                            # fallback
                            logger.warning(
                                f"Unable to automatically determine content_type for {source_name}"
                            )
                            content_type = "text/html; charset=utf8"

                    if not isinstance(self.store_backends["static_assets"], GXCloudStoreBackend):
                        self.store_backends["static_assets"].set(
                            store_key,
                            f.read(),
                            content_encoding=content_encoding,
                            content_type=content_type,
                        )

    def _unzip_assets(self, assets_full_path: str, unzip_directory: str) -> bool:
        """
        This function receives an `assets_full_path` parameter,
        (e.g. "/home/joe/libs/my_python_libs.zip/great_expectations/render/view/static")
        and an `unzip_directory` parameter (e.g. "/tmp/extract_statics_here")

        If `assets_full_path` is a folder inside a zip, then said folder is extracted
        (unzipped) to the `unzip_directory` and this function returns True.
        Otherwise, this function returns False
        """

        static_assets_source_absdir = os.path.abspath(assets_full_path)  # noqa: PTH100 # FIXME CoP

        zip_re = re.match(
            f"(.+[.]zip){re.escape(os.sep)}(.+)",
            static_assets_source_absdir,
            flags=re.IGNORECASE,
        )

        if zip_re:
            zip_filename = zip_re.groups()[0]  # e.g.: /home/joe/libs/my_python_libs.zip
            path_in_zip = zip_re.groups()[1]  # great_expectations/render/view/static
            if is_zipfile(zip_filename):
                with ZipFile(zip_filename) as zipfile:
                    static_files_to_extract = [
                        file for file in zipfile.namelist() if file.startswith(path_in_zip)
                    ]
                    zipfile.extractall(unzip_directory, static_files_to_extract)
                return True

        return False

    @property
    def config(self) -> dict:
        return self._config



from unittest.mock import Mock  # noqa: TID251 # FIXME CoP

import pytest

from great_expectations.compatibility.sqlalchemy_compatibility_wrappers import (
    add_dataframe_to_db,
)
from great_expectations.core.batch_spec import SqlAlchemyDatasourceBatchSpec
from great_expectations.execution_engine import SqlAlchemyExecutionEngine
from great_expectations.execution_engine.sqlalchemy_dialect import GXSqlDialect

try:
    sqlalchemy = pytest.importorskip("sqlalchemy")
except ImportError:
    sqlalchemy = None

from great_expectations.execution_engine.sqlalchemy_batch_data import (
    SqlAlchemyBatchData,
)
from tests.sqlalchemy_test_doubles import Dialect, MockSaEngine
from tests.test_utils import get_sqlite_temp_table_names_from_engine

pytestmark = pytest.mark.sqlalchemy_version_compatibility


@pytest.mark.sqlite
def test_instantiation_with_table_name(sqlite_view_engine):
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(
        engine=sqlite_view_engine
    )
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        table_name="test_table",
    )

    # This is a very hacky type check.
    # A better way would be to figure out the proper parent class for dialects within SQLAlchemy
    assert str(type(batch_data.sql_engine_dialect))[:28] == "<class 'sqlalchemy.dialects."

    assert isinstance(batch_data.selectable, sqlalchemy.Table)

    assert batch_data.use_quoted_name is False


@pytest.mark.sqlite
def test_instantiation_with_query(sqlite_view_engine, test_df):
    add_dataframe_to_db(df=test_df, name="test_table_0", con=sqlite_view_engine)

    query: str = "SELECT * FROM test_table_0"
    # If create_temp_table=False, a new temp table should NOT be created
    # noinspection PyUnusedLocal
    SqlAlchemyBatchData(
        execution_engine=sqlite_view_engine,
        query=query,
        create_temp_table=False,
    )
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 1


# REMOVING PENDING READ OF table.head metric
# def test_head(sqlite_view_engine):
#     # Create a larger table so that we can downsample meaningfully
#     df = pd.DataFrame({"a": range(100)})
#     df.to_sql(name="test_table_2", con=sqlite_view_engine, index=False)
#
#     engine = SqlAlchemyExecutionEngine(engine=sqlite_view_engine)
#     batch_data = SqlAlchemyBatchData(
#         execution_engine=engine,
#         table_name="test_table_2",
#     )
#     engine.load_batch_data("__", batch_data)
#     validator = Validator(execution_engine=engine)
#     df = validator.head()
#     assert df.shape == (5, 2)
#
#     assert validator.head(fetch_all=True).shape == (100, 2)
#     assert validator.head(n_rows=20).shape == (20, 2)
#     assert validator.head(n_rows=20, fetch_all=True).shape == (100, 2)


@pytest.mark.sqlite
def test_instantiation_with_and_without_temp_table(sqlite_view_engine, sa):
    print(get_sqlite_temp_table_names_from_engine(sqlite_view_engine))
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 1
    assert get_sqlite_temp_table_names_from_engine(sqlite_view_engine) == {"test_temp_view"}

    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(
        engine=sqlite_view_engine
    )
    # When the SqlAlchemyBatchData object is based on a table, a new temp table is NOT created, even if create_temp_table=True  # noqa: E501 # FIXME CoP
    SqlAlchemyBatchData(
        execution_engine=execution_engine,
        table_name="test_table",
        create_temp_table=True,
    )
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 1

    selectable = sa.select("*").select_from(sa.text("main.test_table"))

    # If create_temp_table=False, a new temp table should NOT be created
    SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=selectable,
        create_temp_table=False,
    )
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 1

    # If create_temp_table=True, a new temp table should be created
    SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=selectable,
        create_temp_table=True,
    )
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 2

    # If create_temp_table=True, a new temp table should be created
    SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=selectable,
        # create_temp_table defaults to True
    )
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 3

    # testing whether schema is supported
    selectable = sa.select("*").select_from(sa.table(name="test_table", schema="main"))
    SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=selectable,
        # create_temp_table defaults to True
    )
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 4

    # test schema with execution engine
    # TODO : Will20210222 Add tests for specifying schema with non-sqlite backend that actually supports new schema creation  # noqa: E501 # FIXME CoP
    my_batch_spec = SqlAlchemyDatasourceBatchSpec(
        **{
            "table_name": "test_table",
            "batch_identifiers": {},
            "schema_name": "main",
        }
    )
    res = execution_engine.get_batch_data_and_markers(batch_spec=my_batch_spec)
    assert len(res) == 2


@pytest.mark.sqlite
def test_instantiation_with_unknown_dialect(sqlite_view_engine):
    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(
        engine=sqlite_view_engine
    )
    execution_engine.engine.dialect.name = "not_a_supported_dialect"
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        table_name="test_table",
    )

    assert batch_data.dialect == GXSqlDialect.OTHER


@pytest.mark.sqlite
def test_instantiation_with_temp_table_schema():
    # not supported
    engine = MockSaEngine(dialect=Dialect(dialect="sqlite"))
    execution_engine = Mock(spec=SqlAlchemyExecutionEngine, engine=engine)
    batch_data = SqlAlchemyBatchData(
        execution_engine=execution_engine,
        query="test_query",
        create_temp_table=True,
        temp_table_schema_name="test_schema",
    )
    (query_to_create_temp_table, _temp_table_name) = batch_data._create_temporary_table(
        dialect=GXSqlDialect.SQLITE,
        query="test_query",
        temp_table_schema_name="test_schema",
    )
    assert "test_schema" not in query_to_create_temp_table

    # supported
    for dialect in ["bigquery", "snowflake", "vertica"]:
        engine.dialect.name = dialect
        batch_data = SqlAlchemyBatchData(
            execution_engine=execution_engine,
            query="test_query",
            create_temp_table=True,
            temp_table_schema_name="test_schema",
        )
        (
            query_to_create_temp_table,
            _temp_table_name,
        ) = batch_data._create_temporary_table(
            dialect=GXSqlDialect.SQLITE,
            query="test_query",
            temp_table_schema_name="test_schema",
        )
        assert "test_schema" in query_to_create_temp_table


@pytest.mark.sqlite
def test_instantiation_with_selectable_only_and_no_temp_table(sqlite_view_engine, sa):
    """
    What does this test and why?

    In cases where we create a validator but explicitly set `create_temp_table`=False, we directly use the
    selectable created by SqlAlchemyExecutionEngine's _build_selectable_from_batch_spec() method.
    """  # noqa: E501 # FIXME CoP

    selectable = sa.select("*").select_from(sa.text("main.test_table"))
    # only have the view that is created by the `sqlite_view_engine` fixture
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 1

    execution_engine: SqlAlchemyExecutionEngine = SqlAlchemyExecutionEngine(
        engine=sqlite_view_engine
    )
    SqlAlchemyBatchData(
        execution_engine=execution_engine,
        selectable=selectable,
        create_temp_table=False,
    )
    # No new views were created
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 1

    SqlAlchemyBatchData(
        execution_engine=execution_engine, selectable=selectable, create_temp_table=True
    )
    # One new temp_table was created
    assert len(get_sqlite_temp_table_names_from_engine(sqlite_view_engine)) == 2



from __future__ import annotations

import logging
from string import Formatter
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Tuple, Type, Union

from great_expectations.compatibility import pydantic
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # FIXME CoP
)
from great_expectations.expectations.expectation import (
    BatchExpectation,
    render_suite_parameter_string,
)
from great_expectations.expectations.metadata_types import DataQualityIssues, SupportedDataSources
from great_expectations.render import (
    AtomicDiagnosticRendererType,
    RenderedAtomicContent,
    renderedAtomicValueSchema,
)
from great_expectations.render.components import LegacyRendererType, RenderedStringTemplateContent
from great_expectations.render.renderer.renderer import renderer
from great_expectations.render.renderer_configuration import (
    CodeBlock,
    CodeBlockLanguage,
    RendererConfiguration,
    RendererValueType,
)
from great_expectations.render.util import substitute_none_for_missing

if TYPE_CHECKING:
    from great_expectations.core import ExpectationValidationResult
    from great_expectations.execution_engine import ExecutionEngine
    from great_expectations.expectations.expectation_configuration import ExpectationConfiguration


logger = logging.getLogger(__name__)


EXPECTATION_SHORT_DESCRIPTION = (
    "This Expectation will fail validation if the query returns one or more rows. "
    "The WHERE clause defines the fail criteria."
)
UNEXPECTED_ROWS_QUERY_DESCRIPTION = "A SQL or Spark-SQL query to be executed for validation."
SUPPORTED_DATA_SOURCES = [
    SupportedDataSources.SPARK.value,
    SupportedDataSources.POSTGRESQL.value,
    SupportedDataSources.REDSHIFT.value,
    SupportedDataSources.MYSQL.value,
    SupportedDataSources.BIGQUERY.value,
    SupportedDataSources.SNOWFLAKE.value,
    SupportedDataSources.DATABRICKS.value,
]
DATA_QUALITY_ISSUES = [DataQualityIssues.SQL.value]


class UnexpectedRowsExpectation(BatchExpectation):
    __doc__ = f"""{EXPECTATION_SHORT_DESCRIPTION}

    UnexpectedRowsExpectations facilitate the execution of SQL or Spark-SQL queries \
    as the core logic for an Expectation. UnexpectedRowsExpectations must implement \
    a `_validate(...)` method containing logic for determining whether data returned \
    by the executed query is successfully validated. One is written by default, but \
    can be overridden.

    A successful validation is one where the unexpected_rows_query returns no rows.

    UnexpectedRowsExpectation is a \
    [Batch Expectation](https://docs.greatexpectations.io/docs/guides/expectations/creating_custom_expectations/how_to_create_custom_batch_expectations).

    BatchExpectations are one of the most common types of Expectation.
    They are evaluated for an entire Batch, and answer a semantic question about the Batch itself.

    Args:
        unexpected_rows_query (str): {UNEXPECTED_ROWS_QUERY_DESCRIPTION}

    Returns:
        An [ExpectationSuiteValidationResult](https://docs.greatexpectations.io/docs/terms/validation_result)

    Supported Data Sources:
        [{SUPPORTED_DATA_SOURCES[0]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[1]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[2]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[3]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[4]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[5]}](https://docs.greatexpectations.io/docs/application_integration_support/)
        [{SUPPORTED_DATA_SOURCES[6]}](https://docs.greatexpectations.io/docs/application_integration_support/)
    Data Quality Issues:
        {DATA_QUALITY_ISSUES[0]}
    """

    unexpected_rows_query: Union[str, SuiteParameterDict] = pydantic.Field(
        description=UNEXPECTED_ROWS_QUERY_DESCRIPTION
    )

    metric_dependencies: ClassVar[Tuple[str, ...]] = (
        "unexpected_rows_query.table",
        "unexpected_rows_query.row_count",
    )
    success_keys: ClassVar[Tuple[str, ...]] = ("unexpected_rows_query",)
    domain_keys: ClassVar[Tuple[str, ...]] = (
        "batch_id",
        "row_condition",
        "condition_parser",
    )

    @pydantic.validator("unexpected_rows_query")
    def _validate_query(
        cls, query: Union[str, SuiteParameterDict]
    ) -> Union[str, SuiteParameterDict]:
        if isinstance(query, SuiteParameterDict):
            return query

        parsed_fields = [f[1] for f in Formatter().parse(query)]
        if "batch" not in parsed_fields:
            batch_warning_message = (
                "unexpected_rows_query should contain the {batch} parameter. "
                "Otherwise data outside the configured batch will be queried."
            )
            # instead of raising a disruptive warning, we print and log info
            # in order to make the user aware of the potential for querying
            # data outside the configured batch
            print(batch_warning_message)
            logger.info(batch_warning_message)

        return query.rstrip("; \t\r\n\v\f")

    class Config:
        title = "Custom Expectation with SQL"

        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[UnexpectedRowsExpectation]) -> None:
            BatchExpectation.Config.schema_extra(schema, model)
            schema["properties"]["metadata"]["properties"].update(
                {
                    "data_quality_issues": {
                        "title": "Data Quality Issues",
                        "type": "array",
                        "const": DATA_QUALITY_ISSUES,
                    },
                    "short_description": {
                        "title": "Short Description",
                        "type": "string",
                        "const": EXPECTATION_SHORT_DESCRIPTION,
                    },
                    "supported_data_sources": {
                        "title": "Supported Data Sources",
                        "type": "array",
                        "const": SUPPORTED_DATA_SOURCES,
                    },
                }
            )

    @classmethod
    @override
    def _prescriptive_template(
        cls,
        renderer_configuration: RendererConfiguration,
    ) -> RendererConfiguration:
        renderer_configuration.add_param(
            name="unexpected_rows_query", param_type=RendererValueType.STRING
        )
        renderer_configuration.code_block = CodeBlock(
            code_template_str="$unexpected_rows_query",
            language=CodeBlockLanguage.SQL,
        )
        return renderer_configuration

    @classmethod
    @renderer(renderer_type=LegacyRendererType.PRESCRIPTIVE)
    @render_suite_parameter_string
    @override
    def _prescriptive_renderer(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
        **kwargs,
    ) -> list[RenderedStringTemplateContent]:
        runtime_configuration = runtime_configuration or {}
        styling = runtime_configuration.get("styling")
        params = substitute_none_for_missing(
            configuration.kwargs,  # type: ignore[union-attr] # FIXME CoP
            ["unexpected_rows_query"],
        )

        template_str = "Unexpected rows query: $unexpected_rows_query"

        return [
            RenderedStringTemplateContent(
                content_block_type="string_template",
                string_template={
                    "template": template_str,
                    "params": params,
                    "styling": styling,
                },
            )
        ]

    @classmethod
    @renderer(renderer_type=AtomicDiagnosticRendererType.OBSERVED_VALUE)
    @override
    def _atomic_diagnostic_observed_value(
        cls,
        configuration: Optional[ExpectationConfiguration] = None,
        result: Optional[ExpectationValidationResult] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> RenderedAtomicContent:
        renderer_configuration: RendererConfiguration = RendererConfiguration(
            configuration=configuration,
            result=result,
            runtime_configuration=runtime_configuration,
        )

        unexpected_row_count = (
            result.get("result").get("observed_value") if result is not None else None
        )

        template_str = ""
        if isinstance(unexpected_row_count, (int, float)):
            renderer_configuration.add_param(
                name="observed_value",
                param_type=RendererValueType.NUMBER,
                value=unexpected_row_count,
            )

            template_str = "$observed_value unexpected "
            if unexpected_row_count == 1:
                template_str += "row"
            else:
                template_str += "rows"

        renderer_configuration.template_str = template_str

        value_obj = renderedAtomicValueSchema.load(
            {
                "template": renderer_configuration.template_str,
                "params": renderer_configuration.params.dict(),
                "meta_notes": renderer_configuration.meta_notes,
                "schema": {"type": "com.superconductive.rendered.string"},
            }
        )
        return RenderedAtomicContent(
            name=AtomicDiagnosticRendererType.OBSERVED_VALUE,
            value=value_obj,
            value_type="StringValueType",
        )

    @override
    def _validate(
        self,
        metrics: dict,
        runtime_configuration: dict | None = None,
        execution_engine: ExecutionEngine | None = None,
    ) -> Union[ExpectationValidationResult, dict]:
        metric_value = metrics["unexpected_rows_query.table"]
        unexpected_row_count = metrics["unexpected_rows_query.row_count"]
        return {
            "success": unexpected_row_count == 0,
            "result": {
                "observed_value": unexpected_row_count,
                "details": {"unexpected_rows": metric_value},
            },
        }



from __future__ import annotations

from great_expectations.analytics.base_event import Action

DOMAIN_OBJECT_ALL_DESERIALIZE_ERROR: Action = Action("domain_object.all_deserialize_error")
DATA_CONTEXT_INITIALIZED: Action = Action("data_context.initialized")
CHECKPOINT_CREATED: Action = Action("checkpoint.created")
CHECKPOINT_DELETED: Action = Action("checkpoint.deleted")
CHECKPOINT_RAN: Action = Action("checkpoint.ran")
EXPECTATION_SUITE_CREATED: Action = Action("expectation_suite.created")
EXPECTATION_SUITE_DELETED: Action = Action("expectation_suite.deleted")
EXPECTATION_SUITE_EXPECTATION_CREATED: Action = Action("expectation_suite.expectation_created")
EXPECTATION_SUITE_EXPECTATION_DELETED: Action = Action("expectation_suite.expectation_deleted")
EXPECTATION_SUITE_EXPECTATION_UPDATED: Action = Action("expectation_suite.expectation_updated")
VALIDATION_DEFINITION_CREATED: Action = Action("validation_definition.created")
VALIDATION_DEFINITION_DELETED: Action = Action("validation_definition.deleted")
VALIDATION_DEFINITION_RAN: Action = Action("validation_definition.ran")
NOTIFICATION_ACTION_RAN: Action = Action("notification.ran")



import re
from unittest import mock
from unittest.mock import ANY
from unittest.mock import ANY as ANY_TEST_ARG

import pytest
from pytest_mock import MockerFixture

from great_expectations.analytics.events import (
    CheckpointCreatedEvent,
    CheckpointDeletedEvent,
    DomainObjectAllDeserializationEvent,
)
from great_expectations.checkpoint.checkpoint import Checkpoint
from great_expectations.core.expectation_suite import ExpectationSuite
from great_expectations.core.factory.checkpoint_factory import CheckpointFactory
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.abstract_data_context import AbstractDataContext
from great_expectations.data_context.store.checkpoint_store import (
    CheckpointStore,
)
from great_expectations.exceptions import DataContextError


@pytest.mark.unit
def test_checkpoint_factory_get_uses_store_get(
    mocker: MockerFixture,
    arbitrary_validation_definition: ValidationDefinition,
):
    # Arrange
    name = "test-checkpoint"
    store = mocker.MagicMock(spec=CheckpointStore)
    store.has_key.return_value = True
    key = store.get_key.return_value
    checkpoint = Checkpoint(name=name, validation_definitions=[arbitrary_validation_definition])
    store.get.return_value = checkpoint
    factory = CheckpointFactory(store=store)

    # Act
    result = factory.get(name=name)

    # Assert
    store.get.assert_called_once_with(key=key)

    assert result == checkpoint


@pytest.mark.unit
def test_checkpoint_factory_get_raises_error_on_missing_key(
    mocker: MockerFixture,
    arbitrary_validation_definition: ValidationDefinition,
):
    # Arrange
    name = "test-checkpoint"
    store = mocker.MagicMock(spec=CheckpointStore)
    store.has_key.return_value = False
    checkpoint = Checkpoint(name=name, validation_definitions=[arbitrary_validation_definition])
    store.get.return_value = checkpoint
    factory = CheckpointFactory(store=store)

    # Act
    with pytest.raises(DataContextError, match=f"Checkpoint with name {name} was not found."):
        factory.get(name=name)

    # Assert
    store.get.assert_not_called()


@pytest.mark.unit
def test_checkpoint_factory_add_uses_store_add(
    mocker: MockerFixture, arbitrary_validation_definition: ValidationDefinition
):
    # Arrange
    name = "test-checkpoint"
    store = mocker.MagicMock(spec=CheckpointStore)
    store.has_key.return_value = False
    key = store.get_key.return_value
    store.get.return_value = None
    factory = CheckpointFactory(store=store)
    checkpoint = Checkpoint(name=name, validation_definitions=[arbitrary_validation_definition])
    store.get.return_value = checkpoint

    # Act
    factory.add(checkpoint=checkpoint)

    # Assert
    store.add.assert_called_once_with(key=key, value=checkpoint.dict())


@pytest.mark.unit
def test_checkpoint_factory_add_raises_for_duplicate_key(mocker: MockerFixture):
    # Arrange
    name = "test-checkpoint"
    store = mocker.MagicMock(spec=CheckpointStore)
    store.has_key.return_value = True
    factory = CheckpointFactory(store=store)
    checkpoint = Checkpoint(
        name=name, validation_definitions=[mocker.Mock(spec=ValidationDefinition)]
    )

    # Act
    with pytest.raises(
        DataContextError,
        match=f"Cannot add Checkpoint with name {name} because it already exists.",
    ):
        factory.add(checkpoint=checkpoint)

    # Assert
    store.add.assert_not_called()


@pytest.mark.unit
def test_checkpoint_factory_delete_uses_store_remove_key(mocker: MockerFixture):
    # Arrange
    name = "test-checkpoint"
    store = mocker.Mock(spec=CheckpointStore)
    store.has_key.return_value = True
    key = store.get_key.return_value
    checkpoint = mocker.Mock(spec=Checkpoint, id=None)
    checkpoint.name = name
    store.get.return_value = checkpoint
    factory = CheckpointFactory(store=store)

    # Act
    factory.delete(name=name)

    # Assert
    store.remove_key.assert_called_once_with(
        key=key,
    )


@pytest.mark.unit
def test_checkpoint_factory_delete_raises_for_missing_checkpoint(mocker: MockerFixture):
    # Arrange
    name = "test-checkpoint"
    store = mocker.MagicMock(spec=CheckpointStore)
    store.has_key.return_value = False
    factory = CheckpointFactory(store=store)

    # Act
    with pytest.raises(
        DataContextError,
        match=f"Cannot delete Checkpoint with name {name} because it cannot be found.",
    ):
        factory.delete(name=name)

    # Assert
    store.remove_key.assert_not_called()


@pytest.mark.filesystem
def test_checkpoint_factory_is_initialized_with_context_filesystem(empty_data_context):
    assert isinstance(empty_data_context.checkpoints, CheckpointFactory)


@pytest.mark.cloud
def test_checkpoint_factory_is_initialized_with_context_cloud(empty_cloud_data_context):
    assert isinstance(empty_cloud_data_context.checkpoints, CheckpointFactory)


@pytest.mark.filesystem
def test_checkpoint_factory_add_success_filesystem(empty_data_context):
    _test_checkpoint_factory_add_success(empty_data_context)


@pytest.mark.cloud
def test_checkpoint_factory_add_success_cloud(empty_cloud_context_fluent):
    _test_checkpoint_factory_add_success(empty_cloud_context_fluent)


def _test_checkpoint_factory_add_success(context):
    # Arrange
    name = "test-checkpoint"
    ds = context.data_sources.add_pandas("my_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")
    batch_def = asset.add_batch_definition("my_batch_definition")

    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="validation_def", data=batch_def, suite=suite)
    )

    checkpoint = Checkpoint(
        name=name,
        validation_definitions=[validation_definition],
    )
    with pytest.raises(DataContextError, match=f"Checkpoint with name {name} was not found."):
        context.checkpoints.get(name)

    # Act
    created_checkpoint = context.checkpoints.add(checkpoint=checkpoint)

    # Assert
    assert created_checkpoint == context.checkpoints.get(name=name)


@pytest.mark.filesystem
def test_checkpoint_factory_delete_success_filesystem(empty_data_context):
    _test_checkpoint_factory_delete_success(empty_data_context)


@pytest.mark.cloud
def test_checkpoint_factory_delete_success_cloud(empty_cloud_context_fluent):
    _test_checkpoint_factory_delete_success(empty_cloud_context_fluent)


def _test_checkpoint_factory_delete_success(context):
    # Arrange
    name = "test-checkpoint"
    ds = context.data_sources.add_pandas("my_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")
    batch_def = asset.add_batch_definition("my_batch_definition")

    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition = context.validation_definitions.add(
        ValidationDefinition(name="validation_def", data=batch_def, suite=suite)
    )

    context.checkpoints.add(
        checkpoint=Checkpoint(
            name=name,
            validation_definitions=[validation_definition],
        )
    )

    # Act
    context.checkpoints.delete(name)

    # Assert
    with pytest.raises(
        DataContextError,
        match=f"Checkpoint with name {name} was not found.",
    ):
        context.checkpoints.get(name)


@pytest.mark.parametrize(
    "context_fixture_name",
    [
        pytest.param("empty_cloud_context_fluent", id="cloud", marks=pytest.mark.unit),
        pytest.param("in_memory_runtime_context", id="ephemeral", marks=pytest.mark.unit),
        pytest.param("empty_data_context", id="filesystem", marks=pytest.mark.filesystem),
    ],
)
def test_checkpoint_factory_all(context_fixture_name: str, request: pytest.FixtureRequest):
    context: AbstractDataContext = request.getfixturevalue(context_fixture_name)

    # Arrange
    ds = context.data_sources.add_pandas("my_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")  # type: ignore[arg-type] # FIXME CoP
    batch_def = asset.add_batch_definition("my_batch_definition")

    suite = context.suites.add(ExpectationSuite(name="my_suite"))
    validation_definition_a = context.validation_definitions.add(
        ValidationDefinition(name="val def a", data=batch_def, suite=suite)
    )

    checkpoint_a = context.checkpoints.add(
        Checkpoint(
            name="a",
            validation_definitions=[validation_definition_a],
        )
    )

    validation_definition_b = context.validation_definitions.add(
        ValidationDefinition(name="val def b", data=batch_def, suite=suite)
    )
    checkpoint_b = context.checkpoints.add(
        Checkpoint(
            name="b",
            validation_definitions=[validation_definition_b],
        )
    )

    # Act
    result = context.checkpoints.all()
    result = sorted(result, key=lambda x: x.name)

    # Assert
    assert [r.name for r in result] == [checkpoint_a.name, checkpoint_b.name]
    assert result == [checkpoint_a, checkpoint_b]


@pytest.mark.unit
def test_checkpoint_factory_all_with_bad_config(
    in_memory_runtime_context: AbstractDataContext, mocker: MockerFixture
):
    analytics_submit_mock = mocker.patch(
        "great_expectations.data_context.store.store.submit_analytics_event"
    )

    # Arrange
    context: AbstractDataContext = in_memory_runtime_context
    ds = context.data_sources.add_pandas("my_datasource")
    asset = ds.add_csv_asset("my_asset", "data.csv")  # type: ignore[arg-type] # FIXME CoP
    batch_def = asset.add_batch_definition("my_batch_definition")
    suite = context.suites.add(ExpectationSuite(name="my_suite"))

    checkpoint_1 = context.checkpoints.add(
        Checkpoint(
            name="1",
            validation_definitions=[
                context.validation_definitions.add(
                    ValidationDefinition(name="vd1", data=batch_def, suite=suite)
                )
            ],
        )
    )
    checkpoint_2 = context.checkpoints.add(
        Checkpoint(
            name="2",
            validation_definitions=[
                context.validation_definitions.add(
                    ValidationDefinition(name="vd2", data=batch_def, suite=suite)
                )
            ],
        )
    )
    # Verify our checkpoints are added
    assert sorted(context.checkpoints.all(), key=lambda cp: cp.name) == [checkpoint_1, checkpoint_2]

    # Make checkpoint_2 invalid. Pydantic will validate the object at creation time
    # but we can invalidate via assignment.
    checkpoint_2.id = {}  # type: ignore[assignment] # done intentionally for test
    checkpoint_2.save()

    # Act
    result = context.checkpoints.all()

    # Assert
    assert result == [checkpoint_1]
    analytics_submit_mock.assert_called_once_with(
        DomainObjectAllDeserializationEvent(
            error_type=ANY_TEST_ARG,
            store_name="CheckpointStore",
        )
    )
    analytics_submit_args = analytics_submit_mock.call_args[0][0]
    assert re.match("pydantic.*ValidationError", analytics_submit_args.error_type)


class TestCheckpointFactoryAnalytics:
    @pytest.mark.filesystem
    def test_checkpoint_factory_add_emits_event_filesystem(self, empty_data_context):
        self._test_checkpoint_factory_add_emits_event(empty_data_context)

    @pytest.mark.cloud
    def test_checkpoint_factory_add_emits_event_cloud(self, empty_cloud_context_fluent):
        self._test_checkpoint_factory_add_emits_event(empty_cloud_context_fluent)

    def _test_checkpoint_factory_add_emits_event(self, context):
        # Arrange
        name = "test-checkpoint"
        ds = context.data_sources.add_pandas("my_datasource")
        asset = ds.add_csv_asset("my_asset", "data.csv")
        batch_def = asset.add_batch_definition("my_batch_definition")

        suite = context.suites.add(ExpectationSuite(name="my_suite"))
        validation_definition = context.validation_definitions.add(
            ValidationDefinition(name="validation_def", data=batch_def, suite=suite)
        )

        checkpoint = Checkpoint(
            name=name,
            validation_definitions=[validation_definition],
        )

        # Act
        with mock.patch(
            "great_expectations.core.factory.checkpoint_factory.submit_event", autospec=True
        ) as mock_submit:
            _ = context.checkpoints.add(checkpoint=checkpoint)

        # Assert
        mock_submit.assert_called_once_with(
            event=CheckpointCreatedEvent(
                checkpoint_id=mock.ANY,
                validation_definition_ids=[mock.ANY for _ in checkpoint.validation_definitions],
            )
        )

    @pytest.mark.filesystem
    def test_checkpoint_factory_delete_emits_event_filesystem(self, empty_data_context):
        self._test_checkpoint_factory_delete_emits_event(empty_data_context)

    @pytest.mark.cloud
    def test_checkpoint_factory_delete_emits_event_cloud(self, empty_cloud_context_fluent):
        self._test_checkpoint_factory_delete_emits_event(empty_cloud_context_fluent)

    def _test_checkpoint_factory_delete_emits_event(self, context):
        # Arrange
        name = "test-checkpoint"
        ds = context.data_sources.add_pandas("my_datasource")
        asset = ds.add_csv_asset("my_asset", "data.csv")
        batch_def = asset.add_batch_definition("my_batch_definition")

        suite = context.suites.add(ExpectationSuite(name="my_suite"))
        validation_definition = context.validation_definitions.add(
            ValidationDefinition(name="validation_def", data=batch_def, suite=suite)
        )

        checkpoint = Checkpoint(
            name=name,
            validation_definitions=[validation_definition],
        )
        checkpoint = context.checkpoints.add(checkpoint=checkpoint)

        # Act
        with mock.patch(
            "great_expectations.core.factory.checkpoint_factory.submit_event", autospec=True
        ) as mock_submit:
            context.checkpoints.delete(name=name)

        # Assert
        mock_submit.assert_called_once_with(
            event=CheckpointDeletedEvent(checkpoint_id=checkpoint.id)
        )


class TestCheckpointFactoryAddOrUpdate:
    CHECKPOINT_NAME = "checkpoint A"

    def test_add_empty_new_checkpoint(self, data_context: AbstractDataContext) -> None:
        # arrange
        checkpoint = Checkpoint(name=self.CHECKPOINT_NAME, validation_definitions=[])

        # act
        created_checkpoint = data_context.checkpoints.add_or_update(checkpoint=checkpoint)

        # assert
        assert created_checkpoint.id
        data_context.checkpoints.get(self.CHECKPOINT_NAME)

    def test_add_new_checkpoint_with_validations(self, data_context: AbstractDataContext) -> None:
        # arrange
        batch_def = (
            data_context.data_sources.add_pandas("data source A")
            .add_dataframe_asset("asset A")
            .add_batch_definition_whole_dataframe("batch def A")
        )

        validation_definitions = [
            ValidationDefinition(
                name="val def A",
                data=batch_def,
                suite=ExpectationSuite(name="suite A"),
            ),
            ValidationDefinition(
                name="val def B",
                data=batch_def,
                suite=ExpectationSuite(name="suite B"),
            ),
        ]
        checkpoint = Checkpoint(
            name=self.CHECKPOINT_NAME, validation_definitions=validation_definitions
        )

        # act
        created_checkpoint = data_context.checkpoints.add_or_update(checkpoint=checkpoint)

        # assert
        assert created_checkpoint.id
        assert len(created_checkpoint.validation_definitions) == len(validation_definitions)
        for val_def, created_val_def in zip(
            validation_definitions, created_checkpoint.validation_definitions
        ):
            assert created_val_def.id
            val_def_dict = val_def.dict()
            val_def_dict["id"] = ANY
            assert val_def_dict == created_val_def.dict()

    def test_update_existing_checkpoint_adds_validations(
        self, data_context: AbstractDataContext
    ) -> None:
        # arrange
        data_context.checkpoints.add(
            checkpoint=Checkpoint(name=self.CHECKPOINT_NAME, validation_definitions=[])
        )

        batch_def = (
            data_context.data_sources.add_pandas("data source A")
            .add_dataframe_asset("asset A")
            .add_batch_definition_whole_dataframe("batch def A")
        )
        validation_definitions = [
            ValidationDefinition(
                name="val def A",
                data=batch_def,
                suite=ExpectationSuite(name="suite A"),
            ),
            ValidationDefinition(
                name="val def B",
                data=batch_def,
                suite=ExpectationSuite(name="suite B"),
            ),
        ]
        checkpoint = Checkpoint(
            name=self.CHECKPOINT_NAME, validation_definitions=validation_definitions
        )

        # act
        created_checkpoint = data_context.checkpoints.add_or_update(checkpoint=checkpoint)

        # assert
        assert created_checkpoint.id
        assert len(checkpoint.validation_definitions) == len(validation_definitions)
        for val_def, created_val_def in zip(
            validation_definitions, created_checkpoint.validation_definitions
        ):
            assert created_val_def.id
            val_def_dict = val_def.dict()
            val_def_dict["id"] = ANY
            assert val_def_dict == created_val_def.dict()

    def test_update_existing_checkpoint_updates_validations(
        self, data_context: AbstractDataContext
    ) -> None:
        # arrange
        batch_def = (
            data_context.data_sources.add_pandas("data source A")
            .add_dataframe_asset("asset A")
            .add_batch_definition_whole_dataframe("batch def A")
        )
        existing_suite = data_context.suites.add(ExpectationSuite(name="suite A"))
        existing_val_def = data_context.validation_definitions.add(
            ValidationDefinition(
                name="val def A",
                data=batch_def,
                suite=existing_suite,
            )
        )

        data_context.checkpoints.add(
            checkpoint=Checkpoint(
                name=self.CHECKPOINT_NAME, validation_definitions=[existing_val_def]
            )
        )
        new_suite_name = "suite C"

        # act
        created_checkpoint = data_context.checkpoints.add_or_update(
            checkpoint=Checkpoint(
                name=self.CHECKPOINT_NAME,
                validation_definitions=[
                    ValidationDefinition(
                        name="val def A",
                        data=batch_def,
                        suite=ExpectationSuite(name=new_suite_name),
                    ),
                ],
            )
        )
        # assert
        for val_def in created_checkpoint.validation_definitions:
            assert val_def.suite.name == new_suite_name

    def test_update_existing_checkpoint_deletes_validations(
        self, data_context: AbstractDataContext
    ) -> None:
        # arrange
        batch_def = (
            data_context.data_sources.add_pandas("data source A")
            .add_dataframe_asset("asset A")
            .add_batch_definition_whole_dataframe("batch def A")
        )
        SUITE_NAME = "suite A"
        VALIDATION_DEFINITION_NAME = "val def A"
        existing_suite_1 = data_context.suites.add(ExpectationSuite(name=SUITE_NAME))
        existing_suite_2 = data_context.suites.add(ExpectationSuite(name="suite B"))
        existing_val_def_1 = data_context.validation_definitions.add(
            ValidationDefinition(
                name=VALIDATION_DEFINITION_NAME,
                data=batch_def,
                suite=existing_suite_1,
            ),
        )
        existing_val_def_2 = data_context.validation_definitions.add(
            ValidationDefinition(
                name="val def B",
                data=batch_def,
                suite=existing_suite_2,
            )
        )

        data_context.checkpoints.add(
            checkpoint=Checkpoint(
                name=self.CHECKPOINT_NAME,
                validation_definitions=[existing_val_def_1, existing_val_def_2],
            )
        )

        # act
        created_checkpoint = data_context.checkpoints.add_or_update(
            checkpoint=Checkpoint(
                name=self.CHECKPOINT_NAME,
                validation_definitions=[
                    ValidationDefinition(
                        name="val def A",
                        data=batch_def,
                        suite=ExpectationSuite(name=SUITE_NAME),
                    ),
                ],
            )
        )

        # assert
        assert len(created_checkpoint.validation_definitions) == 1
        assert created_checkpoint.validation_definitions[0].name == VALIDATION_DEFINITION_NAME
        assert created_checkpoint.validation_definitions[0].suite.name == SUITE_NAME

    def test_add_or_update_is_idempotent(self, data_context: AbstractDataContext) -> None:
        # arrange
        batch_def = (
            data_context.data_sources.add_pandas("data source A")
            .add_dataframe_asset("asset A")
            .add_batch_definition_whole_dataframe("batch def A")
        )
        SUITE_NAME = "suite A"
        VALIDATION_DEFINITION_NAME = "val def A"

        # act
        created_checkpoint_1 = data_context.checkpoints.add_or_update(
            checkpoint=Checkpoint(
                name=self.CHECKPOINT_NAME,
                validation_definitions=[
                    ValidationDefinition(
                        name=VALIDATION_DEFINITION_NAME,
                        data=batch_def,
                        suite=ExpectationSuite(name=SUITE_NAME),
                    ),
                ],
            )
        )
        created_checkpoint_2 = data_context.checkpoints.add_or_update(
            checkpoint=Checkpoint(
                name=self.CHECKPOINT_NAME,
                validation_definitions=[
                    ValidationDefinition(
                        name=VALIDATION_DEFINITION_NAME,
                        data=batch_def,
                        suite=ExpectationSuite(name=SUITE_NAME),
                    ),
                ],
            )
        )

        # assert
        assert created_checkpoint_1 == created_checkpoint_2



from __future__ import annotations

"""Warnings (including deprecation).

This module consolidates warnings throughout our codebase into a single place
so that changes e.g. to verbiage are restricted to this file.
"""
import warnings


def warn_pandas_less_than_2_0_and_sqlalchemy_greater_than_or_equal_2_0() -> None:
    """Warning to emit when using pandas less than v1.4.0 with sqlalchemy greater than or equal to 2.0.0."""  # noqa: E501 # FIXME CoP
    warnings.warn(
        """Please be aware that pandas versions below 2.0.0 may have issues when paired with SQLAlchemy 2.0.0 and above when using pandas + sql functionality (like the pandas read_sql reader method).""",  # noqa: E501 # FIXME CoP
        UserWarning,
    )



import os

import pandas as pd
import pytest

from great_expectations.data_context.util import file_relative_path


@pytest.fixture(scope="module")
def titanic_df() -> pd.DataFrame:
    path = file_relative_path(
        __file__,
        os.path.join(  # noqa: PTH118 # FIXME CoP
            "..",
            "..",
            "test_sets",
            "Titanic.csv",
        ),
    )
    df = pd.read_csv(path)
    return df



from __future__ import annotations

import logging
import pathlib
import re

import pytest

from great_expectations.datasource.fluent.data_connector.regex_parser import (
    RegExParser,
)

logger = logging.getLogger(__file__)


@pytest.fixture
def csv_path() -> pathlib.Path:
    relative_path = pathlib.Path(
        "..",
        "..",
        "..",
        "..",
        "test_sets",
        "taxi_yellow_tripdata_samples",
        "yellow_tripdata_sample_2020-03.csv",
    )
    abs_csv_path = pathlib.Path(__file__).parent.joinpath(relative_path).resolve(strict=True)
    return abs_csv_path


@pytest.fixture
def regex_pattern_two_named_groups() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(?P<month>\d{2})\.csv")


@pytest.fixture
def regex_pattern_first_named_group_second_common_group() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(?P<year>\d{4})-(\d{2})\.csv")


@pytest.fixture
def regex_pattern_first_common_group_second_named_group() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(\d{4})-(?P<month>\d{2})\.csv")


@pytest.fixture
def regex_pattern_two_common_groups() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_(\d{4})-(\d{2})\.csv")


@pytest.fixture
def regex_pattern_no_groups() -> re.Pattern:
    return re.compile(r"yellow_tripdata_sample_\d{4}-\d{2}\.csv")


@pytest.mark.unit
def test_regex_pattern_two_named_groups(
    regex_pattern_two_named_groups: re.Pattern, csv_path: pathlib.Path
):
    regex_pattern: re.Pattern = regex_pattern_two_named_groups
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {
        "year": 1,
        "month": 2,
    }
    assert regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings() == (
        {"year": 1, "month": 2},
        {1: "year", 2: "month"},
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "year": 1,
        "month": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "year",
        2: "month",
    }
    assert regex_parser.group_names() == ["year", "month"]


@pytest.mark.unit
def test_regex_pattern_first_named_group_second_common_group(
    regex_pattern_first_named_group_second_common_group: re.Pattern,
    csv_path: pathlib.Path,
):
    regex_pattern: re.Pattern = regex_pattern_first_named_group_second_common_group
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {"year": 1}
    assert regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings() == (
        {"year": 1, "batch_request_param_2": 2},
        {1: "year", 2: "batch_request_param_2"},
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "year": 1,
        "batch_request_param_2": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "year",
        2: "batch_request_param_2",
    }
    assert regex_parser.group_names() == ["year", "batch_request_param_2"]


@pytest.mark.unit
def test_regex_pattern_first_common_group_second_named_group(
    regex_pattern_first_common_group_second_named_group: re.Pattern,
    csv_path: pathlib.Path,
):
    regex_pattern: re.Pattern = regex_pattern_first_common_group_second_named_group
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {"month": 2}
    assert regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings() == (
        {"batch_request_param_1": 1, "month": 2},
        {1: "batch_request_param_1", 2: "month"},
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "batch_request_param_1": 1,
        "month": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "batch_request_param_1",
        2: "month",
    }
    assert regex_parser.group_names() == ["batch_request_param_1", "month"]


@pytest.mark.unit
def test_regex_pattern_two_common_groups(
    regex_pattern_two_common_groups: re.Pattern, csv_path: pathlib.Path
):
    regex_pattern: re.Pattern = regex_pattern_two_common_groups
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 2
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {}
    assert regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings() == (
        {"batch_request_param_1": 1, "batch_request_param_2": 2},
        {1: "batch_request_param_1", 2: "batch_request_param_2"},
    )
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {
        "batch_request_param_1": 1,
        "batch_request_param_2": 2,
    }
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {
        1: "batch_request_param_1",
        2: "batch_request_param_2",
    }
    assert regex_parser.group_names() == [
        "batch_request_param_1",
        "batch_request_param_2",
    ]


@pytest.mark.unit
def test_regex_pattern_no_groups(regex_pattern_no_groups: re.Pattern, csv_path: pathlib.Path):
    regex_pattern: re.Pattern = regex_pattern_no_groups
    regex_parser = RegExParser(
        regex_pattern=regex_pattern, unnamed_regex_group_prefix="batch_request_param_"
    )

    assert regex_parser.get_num_all_matched_group_values() == 0
    assert regex_parser.get_named_group_name_to_group_index_mapping() == {}
    assert regex_parser.get_all_group_names_to_group_indexes_bidirectional_mappings() == ({}, {})
    assert regex_parser.get_all_group_name_to_group_index_mapping() == {}
    assert regex_parser.get_all_group_index_to_group_name_mapping() == {}
    assert regex_parser.group_names() == []
